#!/usr/bin/env python
"""This script estimates :term:`P-site offsets <P-site offset>`, stratified by read length,
in a :term:`ribosome profiling` dataset. To do so, read alignments are mapped to
their fiveprime ends, and a :term:`metagene` profile surrounding the start 
codon is calculated separately for each read length.

The start codon peak for each read length is heuristically identified as the
largest peak upstream of the start codon, or within a region defined by the user.
The distance between that peak and the start codon itself is taken to be the
:term:`P-site offset` for that read length.


Notes
------
Generate an *ROI file first*
    This script requires an ROI file of :term:`maximal spanning windows <maximal spanning window>`
    surrounding each gene's start codon. This file can be generated by the
    ``generate`` subprogram of the |metagene| script.
    
Check the data
    Users should examine the graphical output to make sure the P-site estimates
    are reasonable, because if clear start codon peaks are not present in the
    data, the algorithm described above will have trouble.

For RNase I only
    This algorithm presumes that the RNase used to prepare the ribosome-protected
    footprints has no appreciable cutting bias, so that footprints may be
    clearly resolved to the edge of the ribosome.


Output files
------------
    OUTBASE_p_offsets.txt
        Tab-delimited text file with two columns. The first is read length,
        and the second the offset from the fiveprime end of that read length
        to the ribosomal P-site. This table can be supplied as the argument 
        for ``--offset`` when using ``--fiveprime_variable`` mapping in any
        of the other scripts in :obj:`plastid.bin`

    OUTBASE_p_offsets.[svg | png | pdf | et c]
        Plot of metagene profiles for each read length, when reads are mapped
        to their 5' ends, :term:`P-site offsets <P-site offset>` are applied.

    OUTBASE_metagene_profiles.txt
        Metagene profiles, stratified by read length, before :term:`P-site offsets <P-site offset>`
        are applied.

    OUTBASE_K_rawcounts.txt
        Saved if ``--keep`` is given on command line. Raw count vectors for each
        :term:`metagene` window specified in input ROI file, without P-site
        mapping rules applied, for reads of length `K`

    OUTBASE_K_normcounts.txt
        Saved if ``--keep`` is given on command line. Normalized count vectors
        for each metagene window specified in input ROI file, without P-site
        mapping rules applied, for reads of length `K`

where `OUTBASE` is supplied by the user.
"""
import sys
import argparse
import inspect
import warnings

import matplotlib
matplotlib.use("Agg")
import numpy
import pandas as pd
import matplotlib.pyplot as plt

from collections import OrderedDict
from plastid.util.scriptlib.argparsers import (
    AlignmentParser,
    PlottingParser,
    BaseParser,
)
from plastid.genomics.roitools import SegmentChain
from plastid.util.io.openers import get_short_name, argsopener, NullWriter, opener
from plastid.util.io.filters import NameDateWriter
from plastid.util.scriptlib.help_formatters import format_module_docstring
from plastid.util.services.exceptions import ArgumentWarning

# to handle deprecated command-line arguments
from plastid.bin.metagene import _get_norm_region

warnings.simplefilter("once")
printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))


def do_count(
        roi_table,
        ga,
        norm_start,
        norm_end,
        min_counts,
        min_len,
        max_len,
        aggregate=False,
        printer=NullWriter()
):
    """Calculate a :term:`metagene profile` for each read length in the dataset
    
    Parameters
    ----------
    roi_table : :class:`pandas.DataFrame`
        Table specifying regions of interest, generated
        by :py:func:`plastid.bin.metagene.do_generate`
    
    ga : |BAMGenomeArray|
        Count data
    
    norm_start : int
        Coordinate in window specifying normalization region start
    
    norm_end : int
        Coordinate in window specifying normalization region end
    
    min_counts : float
        Minimum number of counts in `window[norm_start:norm_end]`
        required for inclusion in metagene profile

    min_len : int
        Minimum read length to include
    
    max_len : int
        Maximum read length to include

    aggregate : bool, optional
        Estimate P-site from aggregate reads at each position, instead of median
        normalized read density. Potentially noisier, but helpful for lower-count
        data or read lengths with few counts. (Default: False)
                             
    printer : file-like, optional
        filehandle to write logging info to (Default: :func:`~plastid.util.io.openers.NullWriter`)
        
               
    Returns
    -------
    dict
        Dictionary of :class:`numpy.ndarray` s of raw counts at each position (column)
        for each window (row)
    
    dict
        Dictionary of :class:`numpy.ndarray` s of normalized counts at each position (column)
        for each window (row), normalized by the total number of counts in that row
        from `norm_start` to `norm_end`
    
    :class:`pandas.DataFrame`
        Metagene profile of median normalized counts at each position across
        all windows, and the number of windows included in the calculation of each
        median, stratified by read length
    """
    window_size = roi_table["window_size"][0]
    upstream_flank = roi_table["zero_point"][0]

    raw_count_dict = OrderedDict()
    norm_count_dict = OrderedDict()
    shape = (len(roi_table), window_size)
    for i in range(min_len, max_len + 1):
        # mask all by default
        raw_count_dict[i] = numpy.ma.MaskedArray(
            numpy.tile(numpy.nan, shape), mask=numpy.tile(True, shape), dtype=float
        )

    for i, row in roi_table.iterrows():
        if i % 1000 == 0:
            printer.write("Counted %s ROIs ..." % (i + 1))

        roi = SegmentChain.from_str(row["region"])
        mask = SegmentChain.from_str(row["masked"])
        roi.add_masks(*mask)
        valid_mask = roi.get_masked_counts(ga).mask

        offset = int(round((row["alignment_offset"])))
        assert offset + roi.length <= window_size

        count_vectors = {}
        for k in raw_count_dict:
            count_vectors[k] = []

        for seg in roi:
            reads = ga.get_reads(seg)
            read_dict = {}
            for k in raw_count_dict:
                read_dict[k] = []

            for read in filter(lambda x: len(x.positions) in read_dict, reads):
                read_dict[len(read.positions)].append(read)

            for k in read_dict:
                count_vector = ga.map_fn(read_dict[k], seg)[1]
                count_vectors[k].extend(count_vector)

        for k in raw_count_dict:
            if roi.strand == "-":
                count_vectors[k] = count_vectors[k][::-1]

            raw_count_dict[k].data[i, offset:offset + roi.length] = numpy.array(count_vectors[k])
            raw_count_dict[k].mask[i, offset:offset + roi.length] = valid_mask

    profile_table = {"x": numpy.arange(-upstream_flank, window_size - upstream_flank)}

    printer.write("Counted %s ROIs total." % (i + 1))
    for k in raw_count_dict:
        k_raw = raw_count_dict[k]

        denominator = numpy.nansum(k_raw[:, norm_start:norm_end], axis=1)
        norm_count_dict[k] = (k_raw.T.astype(float) / denominator).T

        # copy mask from raw counts, then add nans and infs
        norm_counts = numpy.ma.MaskedArray(norm_count_dict[k], mask=k_raw.mask)
        norm_counts.mask[numpy.isnan(norm_counts)] = True
        norm_counts.mask[numpy.isinf(norm_counts)] = True

        with warnings.catch_warnings():
            # ignore numpy mean of empty slice warning, given by numpy in Python 2.7-3.4
            warnings.filterwarnings("ignore", ".*mean of empty.*", RuntimeWarning)
            try:
                if aggregate == False:
                    profile = numpy.ma.median(norm_counts[denominator >= min_counts], axis=0)
                else:
                    profile = numpy.nansum(k_raw[denominator >= min_counts], axis=0)

            # in numpy under Python3.5, this is an IndexError instead of a warning
            except IndexError:
                profile = numpy.zeros_like(profile_table["x"], dtype=float)
            # in new versions of numpy, this is a ValueEror instead of an IndexError
            except ValueError:
                profile = numpy.zeros_like(profile_table["x"], dtype=float)

        num_genes = ((~norm_counts.mask)[denominator >= min_counts]).sum(0)

        profile_table["%s-mers" % k] = profile
        profile_table["%s_regions_counted" % k] = num_genes

    profile_table = pd.DataFrame(profile_table)

    return raw_count_dict, norm_count_dict, profile_table


def main(argv=sys.argv[1:]):
    """Command-line program
    
    Parameters
    ----------
	argv : list, optional
		A list of command-line arguments, which will be processed
		as if the script were called from the command line if
		:py:func:`main` is called directrly.

        Default: `sys.argv[1:]`. The command-line arguments, if the script is
        invoked from the command line
    """
    ap = AlignmentParser(
        allow_mapping=False, input_choices=["BAM"], disabled=[
            "normalize",
            "big_genome",
        ]
    )
    bp = BaseParser()
    alignment_file_parser = ap.get_parser()
    base_parser = bp.get_parser()

    pp = PlottingParser()
    plotting_parser = pp.get_parser()

    parser = argparse.ArgumentParser(
        description=format_module_docstring(__doc__),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        parents=[base_parser, alignment_file_parser, plotting_parser]
    )

    parser.add_argument(
        "--min_counts",
        type=int,
        default=10,
        metavar="N",
        help="Minimum counts required in normalization region " +
        "to be included in metagene average (Default: 10)"
    )
    parser.add_argument(
        "--normalize_over",
        type=int,
        nargs=2,
        metavar="N",
        default=None,
        #default=(20,50),
        help="Portion of each window against which its individual raw count profile" +
        " will be normalized. Specify two integers, in nucleotide" +
        " distance from landmark (negative for upstream, positive for downstream. Surround negative numbers with quotes.). (Default: 20 50)"
    )
    parser.add_argument(
        "--norm_region",
        type=int,
        nargs=2,
        metavar="N",
        default=None,
        help="Deprecated. Use ``--normalize_over`` instead. " +
        "Formerly, Portion of each window against which its individual raw count profile" +
        " will be normalized. Specify two integers, in nucleotide" +
        " distance, from 5\' end of window. (Default: 70 100)"
    )
    parser.add_argument(
        "--require_upstream",
        default=False,
        action="store_true",
        help="If supplied, the P-site offset is taken to be the distance " +
        "between the largest peak upstream of the start codon and " +
        "the start codon itself. Otherwise, the P-site offset is taken " +
        "to be the distance between the largest peak in the entire ROI " +
        "and the start codon. Ignored if ``--constrain`` is used."
    )
    parser.add_argument(
        "--constrain",
        type=int,
        nargs=2,
        default=None,
        metavar="X",
        help="Constrain P-site offset to be between specified distance from " +
        "start codon. Useful for noisy data. " + "(Reasonable set: 10 15; default: not constrained)"
    )
    parser.add_argument(
        "--aggregate",
        default=False,
        action="store_true",
        help="Estimate P-site from aggregate reads at each position, instead " +
        "of median normalized read density. Noisier, but helpful for " +
        "lower-count data or read lengths with few counts. (Default: False)"
    ),
    parser.add_argument(
        "--keep",
        default=False,
        action="store_true",
        help="Save intermediate count files. Useful for additional computations (Default: False)"
    )
    parser.add_argument(
        "--default",
        type=int,
        default=13,
        help=
        "Default 5\' P-site offset for read lengths that are not present or evaluated in the dataset. Unaffected by ``--constrain`` (Default: 13)"
    )

    parser.add_argument(
        "roi_file",
        type=str,
        help="ROI file surrounding start codons, from ``metagene generate`` subprogram"
    )

    parser.add_argument("outbase", type=str, help="Basename for output files")

    # set manual options
    args = parser.parse_args(argv)
    bp.get_base_ops_from_args(args)

    # set defaults
    args.mapping = "fiveprime"
    args.offset = 0
    args.nibble = 0

    # process arguments
    min_len = args.min_length
    max_len = args.max_length
    profiles = max_len + 1 - min_len
    lengths = list(range(min_len, max_len + 1))
    outbase = args.outbase
    title = "Fiveprime read offsets by length" if args.title is None else args.title

    pp.set_style_from_args(args)
    colors = pp.get_colors_from_args(args, profiles)

    printer.write("Opening ROI file %s ..." % args.roi_file)
    with opener(args.roi_file) as roi_fh:
        roi_table = pd.read_table(roi_fh, sep="\t", comment="#", index_col=None, header=0)
        roi_fh.close()

    printer.write("Opening count files %s ..." % ",".join(args.count_files))
    ga = ap.get_genome_array_from_args(args, printer=printer)

    # remove default size filters
    my_filters = ga._filters.keys()
    for f in my_filters:
        ga.remove_filter(f)

    norm_start, norm_end = _get_norm_region(roi_table, args)

    # count
    count_dict, norm_count_dict, metagene_profile = do_count(
        roi_table,
        ga,
        norm_start,
        norm_end,
        args.min_counts,
        min_len,
        max_len,
        aggregate=args.aggregate,
        printer=printer
    )

    # save counts
    profile_fn = "%s_metagene_profiles.txt" % outbase
    with argsopener(profile_fn, args, "w") as metagene_out:
        metagene_profile.to_csv(
            metagene_out,
            sep="\t",
            header=True,
            index=False,
            na_rep="nan",
            columns=["x"] + ["%s-mers" % X for X in lengths]
        )
        metagene_out.close()

    if args.keep == True:
        printer.write("Saving raw and normalized counts ...")
        for k in count_dict:
            count_fn = "%s_%s_rawcounts.txt.gz" % (outbase, k)
            normcount_fn = "%s_%s_normcounts.txt.gz" % (outbase, k)
            mask_fn = "%s_%s_mask.txt.gz" % (outbase, k)
            numpy.savetxt(count_fn, count_dict[k], delimiter="\t")
            numpy.savetxt(normcount_fn, norm_count_dict[k], delimiter="\t")
            numpy.savetxt(mask_fn, norm_count_dict[k].mask, delimiter="\t")

    # plotting & offsets
    printer.write("Plotting and determining offsets ...")
    offset_dict = OrderedDict()

    # Determine scaling factor for plotting metagene profiles
    max_y = numpy.nan
    with warnings.catch_warnings():
        # ignore warnings for slices that contain only NaNs
        warnings.simplefilter("ignore", category=RuntimeWarning)
        for k in lengths:
            max_y = numpy.nanmax([max_y, numpy.nanmax(metagene_profile["%s-mers" % k].values)])

    if numpy.isnan(max_y) or max_y == 0:
        max_y = 1.0

    # parse arguments & set styles
    mplrc = matplotlib.rcParams
    plt_incr = 1.2

    # use this figsize if not specified on command line
    figheight = 1.0 + 0.25 * (profiles - 1) + 0.75 * (profiles)
    default_figsize = (7.5, figheight)

    fig = pp.get_figure_from_args(args, figsize=default_figsize)

    ax = plt.gca()
    plt.title(title)
    plt.xlabel("Distance from CDS start, (nt; 5' end mapping)")
    if args.aggregate == True:
        plt.ylabel("Aggregate read counts (au)")
    else:
        plt.ylabel("Median normalized read density (au)")

    plt.axvline(0.0, color=mplrc["axes.edgecolor"], dashes=[3, 2])

    x = metagene_profile["x"].values
    xmin = x.min()
    xmax = x.max()

    if args.constrain is not None:
        mask = numpy.tile(True, len(x))

        zp = (x == 0).argmax()
        l, r = args.constrain
        if l == r:
            warnings.warn(
                "Minimum and maximum distance constraints are equal (both '%s')." % l,
                ArgumentWarning
            )

        mindist = min(l, r)
        maxdist = max(l, r)

        mask[zp - maxdist:zp - mindist + 1] = False
    elif args.require_upstream == True:
        mask = x >= 0
    else:
        mask = numpy.tile(False, len(x))

    for n, k in enumerate(lengths):
        color = colors[n]
        baseline = plt_incr * n
        y = metagene_profile["%s-mers" % k].values
        #ymask = y[mask]
        ymask = numpy.ma.MaskedArray(y, mask=mask)

        if numpy.isnan(y).all():
            plot_y = numpy.zeros_like(x)
        else:
            if args.aggregate == False:
                plot_y = y / max_y
            else:
                plot_y = y.astype(float) / numpy.nanmax(y) * 0.9

        # plot metagene profiles on common scale, offset by baseline from bottom to top
        ax.plot(x, baseline + plot_y, color=color)
        ax.text(
            xmin,
            baseline,
            "%s-mers" % k,
            ha="left",
            va="bottom",
            color=color,
            transform=matplotlib.transforms.offset_copy(
                ax.transData, fig, x=6.0, y=3.0, units="points",
            )
        )

        ymax = baseline + numpy.nanmax(plot_y)

        # if all valid positions are nan, or if all valid positions are <= 0
        if (~mask).sum() == numpy.isnan(ymask).sum() or numpy.nanmax(ymask) == 0:
            offset = args.default
            usedefault = True
        else:
            offset = -x[numpy.ma.argmax(ymask)]
            usedefault = False

        offset_dict[k] = offset
        if usedefault == False:
            yadj = ymax - 0.2 * plt_incr

            ax.plot([-offset, 0], [yadj, yadj], color=color, dashes=[3, 2])
            ax.text(
                -offset / 2.0,
                yadj,
                "%s nt" % (offset),
                color=color,
                ha="center",
                va="bottom",
                transform=matplotlib.transforms.offset_copy(
                    ax.transData, fig, x=0.0, y=3.0, units="points",
                )
            )

    plt.xlim(xmin, xmax)
    plt.ylim(-0.1, plt_incr + baseline)
    ax.yaxis.set_ticks([])

    # save data as p-site offset table
    fn = "%s_p_offsets.txt" % outbase
    fout = argsopener(fn, args)
    printer.write("Writing offset table to %s ..." % fn)
    fout.write("length\tp_offset\n")
    for k in offset_dict:
        fout.write("%s\t%s\n" % (k, offset_dict[k]))

    fout.write("default\t%s" % args.default)

    fout.close()

    # save plot
    plot_fn = "%s_p_offsets.%s" % (outbase, args.figformat)
    printer.write("Saving plot to %s ..." % plot_fn)
    plt.savefig(plot_fn, dpi=args.dpi, bbox_inches="tight")

    printer.write("Done.")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.readers.wiggle`"""
import unittest
from plastid.util.services.mini2to3 import cStringIO
from nose.plugins.attrib import attr
from plastid.readers.wiggle import WiggleReader

#===============================================================================
# INDEX: test suites
#===============================================================================


@attr(test="unit")
class TestWiggleReader(unittest.TestCase):
    def _do(self, data_str, data_format, expected_results):
        """Execute tests on various formats

        Parameters
        ----------
        data_str : str
            wiggle file data

        data_format : str
            name of expected data format
        """
        reader = WiggleReader(cStringIO.StringIO(data_str))
        for n, (tup1, tup2) in enumerate(zip(expected_results, reader)):
            self.assertEquals(tup1, tup2)

        self.assertEquals(
            n,
            len(expected_results) - 1,
            "Not all results processed: %s vs %s" % (n, len(expected_results))
        )

        self.assertEquals(reader.data_format, data_format)

    def test_read_multispan_multistep_varstep(self):
        self._do(_MULTISPAN_VARSTEP, "variableStep", _MULTISPAN_TUPLES)

    def test_read_multispan_multistep_fixedstep(self):
        self._do(_MULTISPAN_FIXEDSTEP, "fixedStep", _MULTISPAN_TUPLES)

    def test_read_multispan_multistep_bedgraph(self):
        self._do(_MULTISPAN_BEDGRAPH, "bedGraph", _MULTISPAN_TUPLES)


#===============================================================================
# INDEX: test data
#===============================================================================

# expected results from each type of reader
_MULTISPAN_TUPLES = [
    ("chrA", 5, 10, 5),
    ("chrA", 10, 15, 32),
    ("chrA", 25, 30, 40),
    ("chrA", 2000, 2005, 50),
    ("chrA", 2005, 2010, 50.5),
    ("chrA", 2010, 2015, 50.2),
    ("chrB", 0, 20, 1),
    ("chrB", 20, 40, 1),
    ("chrB", 1000, 1020, 30),
    ("chrB", 1020, 1040, 60),
    ("chrB", 1040, 1060, 1),
    ("chrB", 1060, 1062, 5),
    ("chrB", 1070, 1072, 10),
    ("chrB", 1080, 1082, 20),
]

_MULTISPAN_BEDGRAPH = """track type=bedGraph
chrA    5   10  5
chrA    10  15  32
chrA    25  30  40
chrA    2000    2005    50
chrA    2005    2010    50.5
chrA    2010    2015    50.2
chrB    0   20  1
chrB    20  40  1
chrB    1000    1020    30
chrB    1020    1040    60
chrB    1040    1060    1
chrB    1060    1062    5
chrB    1070    1072    10
chrB    1080    1082    20
""".replace("    ", "\t")

_MULTISPAN_VARSTEP = """track type=wiggle_0
variableStep chrom=chrA span=5
6   5
11  32
26  40
2001    50
2006    50.5
2011    50.2
variableStep chrom=chrB span=20
1   1
21  1
1001    30
1021    60
1041    1
variableStep chrom=chrB span=2
1061    5
1071    10
1081    20
""".replace("    ", "\t")

_MULTISPAN_FIXEDSTEP = """track type=wiggle_0
fixedStep chrom=chrA start=6 span=5 step=5
5
32
fixedStep chrom=chrA start=26 span=5 step=5
40
fixedStep chrom=chrA start=2001 span=5 step=5
50
50.5
50.2
fixedStep chrom=chrB span=20 step=20
1
1
fixedStep chrom=chrB start=1001 span=20 step=20
30
60
1
fixedStep chrom=chrB start=1061 span=2 step=10
5
10
20
"""



#!/usr/bin/env python
"""Setup script for plastid. This is boilerplate except we needed to make
some workarounds to accomodate the fact that compiling C extensions requires
importing from cython, numpy, and pysam, which may not be pre-installed.

Specifically, the workaround:

- defines placeholder values for a number of variables (see [PLACEHOLDERS]),
  enabling `pip` to runthe `egg_info` command required for dependency parsing
  and installation. This allows `pip` to install cython, numpy, and pysam
  after a first pass through this file

- re-defines those placeholders if cython, numpy, and pysam are installed
  (see [REDEFINES]), allowing `pip` to build these extensions when it makes
  its second pass through the file

In addition:

- `build_ext`, `install`, and `develop` commands are wrapped to add a
  ``--recythonize`` flag, which causes Cython to regenerate ".c" files from
  ".pyx" files before the commands are run. This is merely a convenience for
  developers to force rebuild even if ".c" files appear up to date (but are not
  e.g. due to a switch in git repo)

- We added a `clean` command class was added to wipe old ".c" files,

"""
__author__ = "Joshua Griffin Dunn"

import glob
import os
import sys
from setuptools import setup, find_packages, Extension, Command
from setuptools.command.install import install
from setuptools.command.develop import develop
from pkg_resources import parse_version

plastid_version = "0.6.1"


#===============================================================================
# Package metadata
#===============================================================================

with open("README.rst") as f:
    long_description = f.read()

setup_requires = [
    "numpy>=1.9.4",
    "pysam>=0.8.4",
    "cython>=0.22.0",
]

packages = find_packages()
packages += [
    "kent",
    "kent.src",
    "kent.src.lib",
#    "kent.src.lib.font",
#    "kent.src.lib.alpha",
#    "kent.src.lib.x86_64",
#    "kent.src.lib.sparc",
#    "kent.src.lib.i386",
#    "kent.src.lib.i686",
#    "kent.src.lib.ppc",
    "kent.src.inc",
    "kent.src.htslib",
    "kent.src.htslib.htslib",
    "kent.src.htslib.cram",
]

kent_dep_filetypes = ["README", "LICENSE", "AUTHORS", "*.c", "*.h", "*.ps", "*.pss"]

package_data = {
    X : kent_dep_filetypes for X in packages
}
package_data[""] = ["*.pyx", "*.pxd"]


# trim dependencies if on readthedocs server, where many dependencies are mocked
on_rtd = os.environ.get("READTHEDOCS", None) == "True"

if not on_rtd:
    install_requires = [
        "scipy>=0.15.1",
        "pandas>=0.17.0",
        "matplotlib>=1.4.0",
        "biopython>=1.64",
        "twobitreader>=3.0.0",
        "termcolor",
    ] + setup_requires
else:
    install_requires = ["cython", "numpy", "pysam", "biopython", "termcolor"]


def get_scripts():
    """Detect command-line scripts automatically

    Returns
    -------
    list
        list of strings describing command-line scripts
    """
    binscripts = [
        X.replace(".py", "") for X in filter(
            lambda x: x.endswith(".py") and  "__init__" not in x,
            os.listdir(os.path.join("plastid",  "bin")),
        )
    ]
    return ["%s = plastid.bin.%s:main" % (X, X) for X in binscripts]



#===============================================================================
#
# Define extension dependencies
#
# Build an extension of portions of Jim Kent's utilities, which are needed by
# BigBedReader and BigWigReader. Kent utilities additionally depend upon
# htslib, which comes bundled with the Kent sources.
#
#===============================================================================

libraries = ["ssl", "crypto", "z"]
base_path = os.getcwd()

kent_htslib = [
    "bgzf.c",
    os.path.join("cram", "cram_codecs.c"),
    os.path.join("cram", "cram_decode.c"),
    os.path.join("cram", "cram_encode.c"),
    os.path.join("cram", "cram_index.c"),
    os.path.join("cram", "cram_io.c"),
    os.path.join("cram", "cram_samtools.c"),
    os.path.join("cram", "cram_stats.c"),
    os.path.join("cram", "files.c"),
    os.path.join("cram", "mFILE.c"),
    os.path.join("cram", "open_trace_file.c"),
    os.path.join("cram", "pooled_alloc.c"),
    os.path.join("cram", "rANS_static.c"),
    os.path.join("cram", "sam_header.c"),
    os.path.join("cram", "string_alloc.c"),
    os.path.join("cram", "thread_pool.c"),
    os.path.join("cram", "vlen.c"),
    os.path.join("cram", "zfio.c"),
    "faidx.c",
    "hfile.c",
    "hfile_net.c",
    "hts.c",
    "knetfile.c",
    "kstring.c",
    "md5.c",
    "sam.c",
    "tbx.c",
]

kent_sources = [
    "aliType.c",
    "asParse.c",
    "base64.c",
    "basicBed.c",
    "bbiAlias.c",
    "bbiRead.c",
    "bbiWrite.c",
    "bigBed.c",
    "binRange.c",
    "bits.c",
    "bPlusTree.c",
    "bwgQuery.c",
    "bwgValsOnChrom.c",
    "cirTree.c",
    "colHash.c",
    "common.c",
    "dlist.c",
    "dnautil.c",
    "dystring.c",
    "errAbort.c",
    "ffAli.c",
    "freeType.c",
    "gemfont.c",
    "hash.c",
    "hex.c",
    "hmmstats.c",
    "htmlColor.c",
    "internet.c",
    "intExp.c",
    "kxTok.c",
    "linefile.c",
    "localmem.c",
    "mgCircle.c",
    "mgPolygon.c",
    "memalloc.c",
    "memgfx.c",
    "obscure.c",
    "osunix.c",
    "pipeline.c",
    "psl.c",
    "rangeTree.c",
    "rbTree.c",
    "sqlNum.c",
    "sqlList.c",
    "tokenizer.c",
    "udc.c",
    "verbose.c",
    "wildcmp.c",
    "zlibFace.c",
]

kent_deps  = [os.path.join(base_path, "kent", "src", "htslib", X) for X in kent_htslib]
kent_deps += [os.path.join(base_path, "kent", "src", "lib", X) for X in kent_sources]


#===============================================================================
# [PLACEHOLDERS]
#
# Placeholders used to build egg_info if C dependencies are not installed.
#
# These placeholders enable `pip` to read this file, identify and install
# dependencies, and then re-run this file _after_ dependencies are installed to
# enable compilation of C extensions.
#
# If dependencies for compiled extensions are installed, these are overwritten
# or extended below in [REDEFINES]
#
#===============================================================================

ext_modules = []
define_macros = []
c_paths = []

include_paths = [
    os.path.join(base_path, "kent", "src", "htslib"),
    os.path.join(base_path, "kent", "src", "inc"),
]

command_classes = {}


# embed method signatures and use Py2 or Py3 char specs, as appropriate
CYTHON_ARGS = {
    "embedsignature": True,
    "language_level": sys.version_info[0],
}

# we add a command class to force rebuild C files from pyx
# and a corresponding command-line argument
CYTHONIZE_COMMAND = "recythonize"  # command-class command from setup.py
CYTHONIZE_ARG = "recythonize"      # '--recythonize', passable to sdist, build_ext et c

# define options accepted by command-line argument
CYTHON_OPTIONS = [
        (CYTHONIZE_ARG,
         None,
         "If supplied,  use Cython to regenerate .c sources from pyx files " +\
         "(default False)"
        ),
]

# turn off --recythonize by default
CYTHON_DEFAULTS = [False]


#===============================================================================
# [REDEFINES]
#
# Detect if dependencies required to build C extensions are pre-installed
# if so, overwrite values from the section above
#===============================================================================

try:
    import numpy
    import pysam
    from Cython.Distutils import build_ext
    from Cython.Distutils.extension import Extension
    from Cython.Build import cythonize

    # Add numpy & Pysam headers to include paths
    # These are structured differently between the two projects,
    # hence the type checking.
    for mod in (numpy, pysam):
        ipart = mod.get_include()
        if isinstance(ipart, str):
            include_paths.append(ipart)
        elif isinstance(ipart, (list, tuple)):
            include_paths.extend(ipart)
        else:
            raise ValueError("Could not parse include path: %s" % ipart)

    define_macros = pysam.get_defines()

    # determine Pysam version
    # Several classes that we cimport moved in pysam 0.10.0,
    # so we define this macro in the build environment
    # to enable map_factories.pyx to cimport from the correct path
    CYTHON_COMPILE_TIME_ENV = {
        "PYSAM10": parse_version(pysam.__version__) >= parse_version("0.10.0")
    }

    # define extensions -------------------------------------------------------

    # redefine extensions
    extension_kwargs = {
        "include_dirs": include_paths,
        "language": "c",
        "cython_directives": CYTHON_ARGS,
    }

    # These extensions have no dependencies on kentUtils,
    # but do have dependencies on pysam
    noinclude_pyx = glob.glob(os.path.join(base_path, "plastid", "genomics", "*.pyx"))
    ext_modules = [
        Extension(
            x.replace(base_path + os.sep, "").replace(".pyx", "").replace(os.sep, "."), [x],
            libraries               = libraries,
            define_macros           = define_macros,
            cython_compile_time_env = CYTHON_COMPILE_TIME_ENV,
            **extension_kwargs
        ) for x in noinclude_pyx
    ] # yapf: disable

    # The following extensions do link to kentUtils, and also zlib
    bbifile = Extension(
        "plastid.readers.bbifile",
        ["plastid/readers/bbifile.pyx"] + kent_deps,
        libraries=libraries,
        **extension_kwargs
    )

    bigwig = Extension(
        "plastid.readers.bigwig",
        ["plastid/readers/bigwig.pyx"] + kent_deps,
        libraries=libraries,
        **extension_kwargs
    )

    bigbed = Extension(
        "plastid.readers.bigbed",
        ["plastid/readers/bigbed.pyx"] + kent_deps,
        libraries=libraries,
        **extension_kwargs
    )

    ext_modules.append(bbifile)
    ext_modules.append(bigwig)
    ext_modules.append(bigbed)

    # define helper functions & classes for build -----------------------------

    # paths to sources
    pyx_paths = []
    for ex in ext_modules:
        pyx_paths.extend([X for X in ex.sources if X.endswith("pyx")])

    c_paths = [X.replace(".pyx", ".c") for X in pyx_paths]

    def wrap_command_classes(baseclass):
        """Add custom command-line `--recythonize` options to
        distutils/setuptools command classes

        Parameters
        ----------
        baseclass : setuptools.Command or subclass (not instance!)
            Command class to modify

        Returns
        -------
        class
            Modified class
        """

        class subclass(baseclass):
            user_options = baseclass.user_options + CYTHON_OPTIONS
            new_options = CYTHON_OPTIONS
            new_defaults = CYTHON_DEFAULTS

            def initialize_options(self):
                baseclass.initialize_options(self)
                for (op, _, _), default in zip(self.new_options, self.new_defaults):
                    setattr(self, op, default)

            def finalize_options(self):
                baseclass.finalize_options(self)
                if "--%s" % CYTHONIZE_ARG in self.distribution.script_args:
                    print("Cythonize on")
                    setattr(self, CYTHONIZE_ARG, True)

            def run(self):
                have_all = all([os.access(X, os.F_OK) for X in c_paths])
                if have_all == False:
                    print("Could not find .c files. Regenerating via recythonize.")
                    setattr(self, CYTHONIZE_ARG, True)
                    self.distribution.script_args.append("--%s" % CYTHONIZE_ARG)

                if getattr(self, CYTHONIZE_ARG) == True:
                    print("Cythonizing")
                    self.run_command(CYTHONIZE_COMMAND)

                return baseclass.run(self)

        subclass.__name__ = "cython_%s" % baseclass.__name__
        return subclass

    class clean_c_files(Command):
        """Remove previously generated .c files"""
        user_options = []
        description = "Remove previously generated .c files"

        def initialize_options(self):
            pass

        def finalize_options(self):
            pass

        def run(self):
            for file_ in c_paths:
                if os.access(file_, os.F_OK):
                    print("clean_c_files: removing %s ..." % file_)
                    os.remove(file_)

    class build_c_from_pyx(build_ext):
        """Regenerate .c files from pyx files if --CYTHONIZE_ARG or
        CYTHONIZE_COMMAND is added to command line
        """
        user_options = build_ext.user_options + CYTHON_OPTIONS
        new_options = CYTHON_OPTIONS
        new_defaults = CYTHON_DEFAULTS

        cython_args = CYTHON_ARGS
        include_path = include_paths
        old_extensions = ext_modules

        description = "Regenerate .c files from .pyx source"

        def initialize_options(self):
            for (op, _, _), default in zip(self.new_options, self.new_defaults):
                setattr(self, op, default)
            build_ext.initialize_options(self)

        def finalize_options(self):
            if "--%s" % CYTHONIZE_ARG in self.distribution.script_args \
              or CYTHONIZE_COMMAND in self.distribution.script_args:

                self.run_command('clean')
                print("build_c_from_pyx: regenerating .c files from Cython")
                extensions = cythonize(
                    ext_modules,
                    compiler_directives=CYTHON_ARGS,
                    compile_time_env=CYTHON_COMPILE_TIME_ENV,
                )
                self.extensions = extensions

            build_ext.finalize_options(self)

        # override so that extensions are only built with build_ext, install, etc
        def run(self):
            pass

    # setup command classes ---------------------------------------------------

    command_classes = {
        CYTHONIZE_COMMAND: build_c_from_pyx,
        'build_ext': wrap_command_classes(build_ext),
        'install': wrap_command_classes(install),
        'develop': wrap_command_classes(develop),
        'clean': clean_c_files,
    }

except ImportError:
    print("plastid: Not all requirements pre-installed. Will need to bootstrap.")


#===============================================================================
# Program body
#===============================================================================

setup(

    name             = "plastid",
    version          = plastid_version,
    author           = "Joshua Griffin Dunn",
    author_email     = "joshua.g.dunn@gmail.com",
    maintainer       = "Joshua Griffin Dunn",
    maintainer_email = "joshua.g.dunn@gmail.com",
    long_description =  long_description,
    long_description_content_type = "text/x-rst",

    description      = "Tools for analysis of genomics & sequencing data",
    license          = "BSD 3-Clause",
    keywords         = "ribosome profiling riboseq rna-seq sequencing genomics biology",
    url              = "https://github.com/joshuagryphon/plastid",
    download_url     = "https://pypi.python.org/pypi/plastid/",
    platforms        = "OS Independent",

    classifiers      = [
         'Development Status :: 5 - Production/Stable',
         'Programming Language :: Python :: 3.6',
         'Programming Language :: Python :: 3.9',

         'Topic :: Scientific/Engineering :: Bio-Informatics',
         'Topic :: Software Development :: Libraries',

         'Intended Audience :: Science/Research',
         'Intended Audience :: Developers',

         'License :: OSI Approved :: BSD License',
         'Operating System :: POSIX',
         'Natural Language :: English',
    ],

    zip_safe = False,
    packages = packages,
    package_data = package_data,

    package_dir = {
        "plastid"  : "plastid",
    },

    entry_points = {
        "console_scripts" : get_scripts()
    },

    ext_modules      = ext_modules,
    cmdclass         = command_classes,

    setup_requires   = setup_requires,
    install_requires = install_requires,

    tests_require    = [ "nose>=1.0" ],
    test_suite       = "nose.collector",

) # yapf: disable



#!/usr/bin/env python
"""This module contains functions for escaping, unescaping, and parsing 
tokens from the ninth column of `GTF2`_ and `GFF3`_ files.

Important methods
-----------------
:py:func:`make_GTF2_tokens`
    Format a dictionary of attributes as `GTF2`_ column 9 attributes

:py:func:`make_GFF3_tokens`
    Format a dictionary of attributes as `GFF3`_ column 9 attributes

:py:func:`parse_GTF2_tokens`
    Parse `GTF2`_ column 9 tokens into a dictionary of key-value pairs

:py:func:`parse_GFF3_tokens`
    Parse `GFF3`_ column 9 tokens into a dictionary of key-value pairs

See also
--------

  - `The Sequence Ontology GFF3 specification <http://www.sequenceontology.org/gff3.shtml>`_
  - `The Brent lab GTF2.2 specification <http://mblab.wustl.edu/GTF22.html>`_
"""

# Unit tests for these are in :py:mod:`plastid.test.unit.genomics.readers.test_gff`

import re
import shlex
import copy
from plastid.util.services.exceptions import FileFormatWarning, warn

gtfpat = re.compile(r"^ *([^ ]*) +(.*) *$")

# From the spec: http://www.sequenceontology.org/gff3.shtml
# In addition to Parent, the Alias, Note, Dbxref and Ontology_term attributes can have multiple values.
# Also, SGD uses 'dbxref' instead of 'Dbxref'
_GFF3_DEFAULT_LISTS = ("Parent", "Alias", "Note", "Dbxref", "Ontology_term", "dbxref")

#===============================================================================
# INDEX: helper functions for escaping
#===============================================================================

# must escape % first, otherwise we'll end up escaping everything else,
# since other escape codes start with percent signs
_GFF3_escape_sequences = [
    ('%', '%25'),  # percent signs MUST be escaped FIRST 
    (';', '%3B'),
    (',', '%2C'),
    ('=', '%3D'),
    ('&', '%26'),
    ('\x00', '%00'),
    ('\x01', '%01'),
    ('\x02', '%02'),
    ('\x03', '%03'),
    ('\x04', '%04'),
    ('\x05', '%05'),
    ('\x06', '%06'),
    ('\x07', '%07'),
    ('\x08', '%08'),
    ('\t', '%09'),
    ('\n', '%0A'),
    ('\x0b', '%0B'),
    ('\x0c', '%0C'),
    ('\r', '%0D'),
    ('\x0e', '%0E'),
    ('\x0f', '%0F'),
    ('\x10', '%10'),
    ('\x11', '%11'),
    ('\x12', '%12'),
    ('\x13', '%13'),
    ('\x14', '%14'),
    ('\x15', '%15'),
    ('\x16', '%16'),
    ('\x17', '%17'),
    ('\x18', '%18'),
    ('\x19', '%19'),
    ('\x1a', '%1A'),
    ('\x1b', '%1B'),
    ('\x1c', '%1C'),
    ('\x1d', '%1D'),
    ('\x1e', '%1E'),
    ('\x1f', '%1F'),
    ('\x7f', '%7F'),
    ('\x80', '%80'),
    ('\x81', '%81'),
    ('\x82', '%82'),
    ('\x83', '%83'),
    ('\x84', '%84'),
    ('\x85', '%85'),
    ('\x86', '%86'),
    ('\x87', '%87'),
    ('\x88', '%88'),
    ('\x89', '%89'),
    ('\x8a', '%8A'),
    ('\x8b', '%8B'),
    ('\x8c', '%8C'),
    ('\x8d', '%8D'),
    ('\x8e', '%8E'),
    ('\x8f', '%8F'),
    ('\x90', '%90'),
    ('\x91', '%91'),
    ('\x92', '%92'),
    ('\x93', '%93'),
    ('\x94', '%94'),
    ('\x95', '%95'),
    ('\x96', '%96'),
    ('\x97', '%97'),
    ('\x98', '%98'),
    ('\x99', '%99'),
    ('\x9a', '%9A'),
    ('\x9b', '%9B'),
    ('\x9c', '%9C'),
    ('\x9d', '%9D'),
    ('\x9e', '%9E'),
    ('\x9f', '%9F')
]
"""List mapping characters to their escape sequences, per the `GFF3`_ specification"""

_GTF2_escape_sequences = copy.deepcopy(_GFF3_escape_sequences)
_GTF2_escape_sequences.append(("\"", "%22"))
"""List mapping characters to their escape sequences for `GTF2`_. These are undefined,
but we are using `GFF3`_ characters plus double quotation marks as a convention.
"""


def escape(inp, char_pairs):
    """Escape reserved characters specified in the list of tuples `char_pairs`
    
    Parameters
    ----------
    inp : str
        Input string
    
    chair_pairs : list
        List of tuples of (character, escape sequence for character)
    
    
    Returns
    -------
    str
        Escaped output
    
    
    See also
    --------
    unescape_GFF3
    """
    for char_, repl in char_pairs:
        inp = inp.replace(char_, repl)

    return inp


def unescape(inp, char_pairs):
    """Unescape reserved characters specified in the list of tuples `char_pairs`
    
    Parameters
    ----------
    inp : str
        Input string
    
    
    Returns
    -------
    str
        Unescaped output
    
    
    See also
    --------
    escape_GFF3
    """
    for repl, char_ in reversed(char_pairs):
        inp = inp.replace(char_, repl)

    return inp


def escape_GFF3(inp):
    """Escape reserved characters in `GFF3`_ tokens using percentage notation.
    
    In the `GFF3`_ spec, reserved characters include:
    
        - control characters (ASCII 0-32, 127, and 128-159)
        
        - tab, newline, & carriage return
        
        - semicolons & commas
        
        - the percent sign
        
        - the equals sign
        
        - the ampersand  
    
    Parameters
    ----------
    inp : str
        Input string
    
    chair_pairs : list
        List of tuples of (character, escape sequence for character)
    
    
    Returns
    -------
    str
        Escaped output
    
    
    See also
    --------
    unescape_GFF3
    """
    return escape(inp, _GFF3_escape_sequences)


def unescape_GFF3(inp):
    """Unescape reserved characters in `GFF3`_ tokens using percentage notation.
    
    In the `GFF3`_ spec, reserved characters include:
    
        - control characters (ASCII 0-32, 127, and 128-159)
        
        - tab, newline, & carriage return
        
        - semicolons & commas
        
        - the percent sign
        
        - the equals sign
        
        - the ampersand  
    
    Parameters
    ----------
    inp : str
        Input string
    
    
    Returns
    -------
    str
        Unescaped output
    
    
    See also
    --------
    escape_GFF3
    """
    return unescape(inp, _GFF3_escape_sequences)


def escape_GTF2(inp):
    """Escape reserved characters in `GTF2`_ tokens using percentage notation.
    While the `GTF2`_ spec is agnostic for escaping, it is useful when adding
    extra attributes to files. As a convention, we escape the characters
    specified in the `GFF3`_ spec, as well as double quotation marks.
    
    In the `GTF2`_ spec, reserved characters include:
    
        - control characters (ASCII 0-32, 127, and 128-159)
        
        - tab, newline, & carriage return
        
        - semicolons & commas
        
        - the percent sign
        
        - the equals sign
        
        - the ampersand  
    
    Parameters
    ----------
    inp : str
        Input string
    
    chair_pairs : list
        List of tuples of (character, escape sequence for character)
    
    
    Returns
    -------
    str
        Escaped output
    
    
    See also
    --------
    unescape_GFF3
    """
    return escape(inp, _GTF2_escape_sequences)


def unescape_GTF2(inp):
    """Unescape reserved characters in `GTF2`_ tokens using percentage notation.
    While the `GTF2`_ spec is agnostic for escaping, it is useful when adding
    extra attributes to files. As a convention, we escape the characters
    specified in the `GFF3`_ spec, as well as single quotation marks.
        
    In the `GFF3`_ spec, reserved characters include:
    
        - control characters (ASCII 0-32, 127, and 128-159)
        
        - tab, newline, & carriage return
        
        - semicolons & commas
        
        - the percent sign
        
        - the equals sign
        
        - the ampersand  
    
    Parameters
    ----------
    inp : str
        Input string
    
    
    Returns
    -------
    str
        Unescaped output
    
    
    See also
    --------
    escape_GFF3
    """
    return unescape(inp, _GTF2_escape_sequences)


#===============================================================================
# INDEX: attribute token formatting and parsing
#===============================================================================


def _make_generic_tokens(attr, excludes=None, join_pat='%s %s; ', escape=None):
    """Helper function to convert the `attr` dict of a |SegmentChain|
    into the string representation used in GFF files. This includes
    URL escaping of keys and values, and catenating lists with `','`
    before string conversion
    
    Parameters
    ----------
    attr : dict
        Dictionary of key-value pairs to export
        
    excludes : list<str>
        List of keys to exclude from string
        
    join_pat
        printf-style pattern explaining how to join key:value pairs
    
    escape : None or func, optional
        If None, no special characters are escaped. If a function, that
        funciton will be used to perform the escaping. (Default: `False`)
        
    Returns
    -------
    str
    """
    f = lambda x: x[0] not in excludes
    if escape is None:
        esc = lambda inp: inp
    else:
        esc = lambda inp: escape(str(inp))

    excludes = [] if excludes is None else excludes

    ltmp = []
    for key, val in filter(f, attr.items()):
        if isinstance(val, list):
            val = ",".join([esc(X) for X in val])
        else:
            val = esc(val)
        ltmp.append(join_pat % (esc(key), val))

    return ''.join(ltmp)


def make_GFF3_tokens(attr, excludes=None, escape=True):
    """Helper function to convert the `attr` dict of a |SegmentChain|
    into the string representation used in `GFF3`_ files. This includes
    URL escaping of special characters, and catenating lists with '`,`'
    before string conversion

    Examples
    --------
        >>> d = {'a':1,'b':2,'c':3,'d':4,'e':5,'z':26,'text':"something; with escape sequences"}
        >>> _make_GFF3_tokens(d)
        'a=1;c=3;b=2;e=5;d=4;z=26;text=something%3B with escape sequences'
    
        >>> excludes=['a','b','c']
        >>> _make_GFF3_tokens(d,excludes)
        'e=5;d=4;z=26;text=something%3B with escape sequences'

        >>> d = {'a':1,'b':2,'c':[3,7],'d':4,'e':5,'z':26}
        >>> _make_GFF3_tokens(d)
        'a=1;c=3,7;b=2;e=5;d=4;z=26'


    Parameters
    ----------
    attr : dict
        Dictionary of key-value pairs to export
        
    excludes : list, optional
        List of keys to exclude from string
        
    escape : bool, optional
        If True, special characters in output are `GFF3`_-escaped (Default: `True`)
        
    Returns
    -------
    str
        Data formatted for *attributes* column of `GFF3`_ (column 9)
    """
    escape = escape_GFF3 if escape == True else None
    excludes = [] if excludes is None else excludes

    return _make_generic_tokens(attr, excludes=excludes, join_pat="%s=%s;", escape=escape)


def make_GTF2_tokens(attr, excludes=None, escape=True):
    """Helper function to convert the `attr` dict  of a |SegmentChain|
    into the string representation used in `GTF2`_ files. By default, special
    characters defined in the `GFF3`_ spec will be URL-escaped.

    Examples
    --------
        >>> d = {'transcript_id' : 't;id', 'a':1,'b':2,'c':3,'d':4,'e':5,'z':26,
                    'gene_id' : 'gid'}
        >>> _make_GTF2_tokens(d)
        'transcript_id "t%3Bid"; gene_id "gid"; a "1"; c "3"; b "2"; e "5"; d "4"; z "26";'
    
        >>> excludes=['a','b','c']
        >>> _make_GTF2_tokens(d,excludes)
        'transcript_id "t%3Bid"; gene_id "gid"; e "5"; d "4"; z "26";'


    Parameters
    ----------
    attr : dict
        Dictionary of key-value pairs to export
        
    excludes : list, optional
        List of keys to exclude from string
        
    escape : bool, optional
        If True, special characters in output are `GTF2`_-escaped (Default: `True`)
        
    Returns
    -------
    str
        Data formatted for *attributes* column of `GTF2`_ (column 9)
    """
    excludes = [] if excludes is None else excludes
    excludes.extend(["transcript_id", "gene_id"])
    stmp = 'gene_id "%s"; transcript_id "%s"; ' % (attr.get("gene_id"), attr.get("transcript_id"))

    if escape == True:
        escape = escape_GTF2
    else:
        escape = None

    return stmp + _make_generic_tokens(
        attr, excludes=excludes, join_pat='%s "%s"; ', escape=escape
    ).strip(" ")


def parse_GFF3_tokens(inp, list_types=None):
    """Helper function to parse tokens in the final column of a `GFF3`_ file
    into a dictionary of attributes. Because, the following attributes are
    permitted to have multiple values in the `GFF3`_ spec, their values, if present
    are returned as lists in the dictionary rather than strings:
    
        - `Parent`
        - `Alias`
        - `Note`
        - `Dbxref`
        - `Ontology_term`
 
    All values are unescaped folowing the `GFF3`_ specification.
 
    Examples
    --------
        >>> tokens = 'a=1;c=3;b=2;e=5;d=4;z=26,Parents=gene01'
        >>> parse_GFF3_tokens(tokens)
        {'a': '1', 'c': '3', 'b': '2', 'e': '5', 'd': '4', 'z': '26', 'parents' : ['gene01'] }

        >>> tokens = 'a=1;c=3,7;b=2;e=5;d=4;z=26,Parents=gene01,gene02'
        >>> parse_GFF3_tokens(tokens)
        {'a': '1', 'c': '3,7', 'b': '2', 'e': '5', 'd': '4', 'z': '26', 'parents' : ['gene01','gene02']}

 
    Parameters
    ----------
    inp : str
        Ninth column of `GFF3`_ entry
    
    list_types : list, optional
        Names of attributes that should be returned as lists
        (Default: %s)
         
    Returns
    -------
    dict : key-value pairs
    """ % ",".join(_GFF3_DEFAULT_LISTS)
    if list_types is None:
        list_types = _GFF3_DEFAULT_LISTS

    d = {}
    items = inp.strip("\n").strip(";").split(";")
    for item in items:
        if len(item) > 0:
            key, val = item.split("=")
            key = unescape_GFF3(key.strip(" "))
            if key in list_types:
                val = [unescape_GFF3(X) for X in val.strip(" ").split(",")]
            else:
                val = unescape_GFF3(val.strip(" "))

            if key in d:
                warn(
                    "Found duplicate attribute key '%s' in GFF3 line. Catenating value with previous value for key in attr dict:\n    %s"
                    % (key, inp), FileFormatWarning
                )
                val = "%s,%s" % (d[key], val)
            d[key] = val
    return d


def parse_GTF2_tokens(inp):
    """Helper function to parse tokens in the final column of a `GTF2`_ file
    into a dictionary of attributes. All attributes are returned as strings,
    and are unescaped if GFF escape sequences (e.g. *'%2B'*) are present.

    If duplicate keys are present (e.g. as in GENCODE `GTF2`_ files),
    their values are catenated, separated by a comma.
    
    Examples
    --------
        >>> tokens = 'gene_id "mygene"; transcript_id "mytranscript";'
        >>> parse_GTF2_tokens(tokens)
        {'gene_id' : 'mygene', 'transcript_id' : 'mytranscript'}
    
        >>> tokens = 'gene_id "mygene"; transcript_id "mytranscript"'
        >>> parse_GTF2_tokens(tokens)
        {'gene_id' : 'mygene', 'transcript_id' : 'mytranscript'}
    
        >>> tokens = 'gene_id "mygene;"; transcript_id "myt;ranscript"'
        >>> parse_GTF2_tokens(tokens)
        {'gene_id' : 'mygene;', 'transcript_id' : 'myt;ranscript'}
    
        >>> tokens = 'gene_id "mygene"; transcript_id "mytranscript"; tag "tag value";'
        >>> parse_GTF2_tokens(tokens)
        {'gene_id' : 'mygene', 'tag' : 'tag value', 'transcript_id' : 'mytranscript'}

        >>> tokens = 'gene_id "mygene"; transcript_id "mytranscript"; tag "tag value"; tag "tag value 2";'
        >>> parse_GTF2_tokens(tokens)
        {'gene_id' : 'mygene', 'tag' : 'tag value,tag value 2', 'transcript_id' : 'mytranscript'}



    Parameters
    ----------
    inp : str
        Ninth column of `GTF2`_ entry
        
    Returns
    -------
    dict : key-value pairs
    """
    d = {}
    items = shlex.split(inp.strip("\n"))
    assert len(items) % 2 == 0
    for i in range(0, len(items), 2):
        key = unescape_GTF2(items[i])
        val = items[i + 1]
        # require separation by semicolons for all but final token
        if i + 1 < len(items) - 2:
            assert val.endswith(";")

        if val.endswith(";"):
            val = val[:-1]

        if key in d:
            warn(
                "Found duplicate attribute key '%s' in GTF2 line. Catenating value with previous value for key in attr dict:\n    %s"
                % (key, inp), FileFormatWarning
            )
            d[key] = "%s,%s" % (d[key], unescape_GTF2(val))

        else:
            d[key] = unescape_GTF2(val)

    return d



#!/usr/bin/env python
"""Estimate :term:`sub-codon phasing` in a :term:`ribosome profiling` dataset,
stratified by read length.

Because ribosomes step three nucleotides in each cycle of translation elongation,
in many :term:`ribosome profiling` datasets a triplet periodicity is observable
in the distribution of :term:`ribosome-protected footprints <footprint>`.

In a good dataset, 70-90% of the reads on a codon fall within the first of the
three codon positions. This allows deduction of translation reading frames, if
the reading frame is not known *a priori.* See :cite:`Ingolia2009` for more
details.

Output files
------------
    OUTBASE_phasing.txt
        Read phasing for each read length

    OUTBASE_phasing.svg
        Plot of phasing by read length

where `OUTBASE` is supplied by the user.

 .. note::

    To avoid double-counting of codons, users should either use an *ROI file*
    made by the ``generate`` subprogram of the |metagene| script, or supply
    an :term:`annotation` file that includes only one transcript isoform per
    gene.

"""
import sys
import argparse
import inspect
import warnings

import pandas as pd
import numpy
import matplotlib
matplotlib.use("Agg")
from plastid.util.scriptlib.argparsers import (
    AlignmentParser,
    AnnotationParser,
    PlottingParser,
    BaseParser,
)
from plastid.util.io.openers import get_short_name, argsopener, read_pl_table
from plastid.util.io.filters import NameDateWriter
from plastid.util.scriptlib.help_formatters import format_module_docstring
from plastid.util.services.exceptions import DataWarning, ArgumentWarning
from plastid.plotting.plots import phase_plot
from plastid.genomics.roitools import SegmentChain

warnings.simplefilter("once")
printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))


def roi_row_to_cds(row):
    """Helper function to extract coding portions from maximal spanning windows
    flanking CDS starts that are created by |metagene| ``generate`` subprogram.
    
    Parameters
    ----------
    row : (int, Series)
        Row from a :class:`pandas.DataFrame` of an ROI file made by the |metagene|
        ``generate`` subprogram
        
    Returns
    -------
    |SegmentChain|
        Coding portion of maximal spanning window
    """
    chainstr, alignment_offset, zero_point = row[1][["region", "alignment_offset", "zero_point"]]
    chain = SegmentChain.from_str(chainstr)
    cds_start = zero_point - alignment_offset
    subchain = chain.get_subchain(cds_start, chain.length)
    return subchain


def main(argv=sys.argv[1:]):
    """Command-line program
    
    Parameters
    ----------
	argv : list, optional
		A list of command-line arguments, which will be processed
		as if the script were called from the command line if
		:py:func:`main` is called directly.

        Default: `sys.argv[1:]`. The command-line arguments, if the script is
        invoked from the command line
    """
    al = AlignmentParser(
        disabled=["normalize", "big_genome", "spliced_bowtie_files"],
        input_choices=["BAM"],
    )
    an = AnnotationParser()
    pp = PlottingParser()
    bp = BaseParser()
    plotting_parser = pp.get_parser()

    alignment_file_parser = al.get_parser(conflict_handler="resolve")
    annotation_file_parser = an.get_parser(conflict_handler="resolve")
    base_parser = bp.get_parser()

    parser = argparse.ArgumentParser(
        description=format_module_docstring(__doc__),
        formatter_class=argparse.RawDescriptionHelpFormatter,
        conflict_handler="resolve",
        parents=[base_parser, annotation_file_parser, alignment_file_parser, plotting_parser]
    )

    parser.add_argument("roi_file",type=str,nargs="?",default=None,
                        help="Optional. ROI file of maximal spanning windows surrounding start codons, "+\
                             "from ``metagene generate`` subprogram. Using this instead of `--annotation_files` "+\
                             "prevents double-counting of codons when multiple transcript isoforms exist "+\
                             "for a gene. See the documentation for `metagene` for more info about ROI files."+\
                             "If an ROI file is not given, supply an annotation with ``--annotation_files``")
    parser.add_argument("outbase", type=str, help="Required. Basename for output files")
    parser.add_argument(
        "--codon_buffer",
        type=int,
        default=5,
        help="Codons before and after start codon to ignore (Default: 5)"
    )

    args = parser.parse_args(argv)
    bp.get_base_ops_from_args(args)
    pp.set_style_from_args(args)
    gnd = al.get_genome_array_from_args(args, printer=printer)

    read_lengths = list(range(args.min_length, args.max_length + 1))
    codon_buffer = args.codon_buffer

    dtmp = {
        "read_length": numpy.array(read_lengths),
        "reads_counted": numpy.zeros_like(read_lengths, dtype=int),
    }

    if args.roi_file is not None:
        using_roi = True
        roi_table = read_pl_table(args.roi_file)
        regions = roi_table.iterrows()
        transform_fn = roi_row_to_cds
        back_buffer = -1
        if len(args.annotation_files) > 0:
            warnings.warn(
                "If an ROI file is given, annotation files are ignored. Pulling regions from '%s'. Ignoring '%s'"
                % (args.roi_file, ", ".join(args.annotation_files)), ArgumentWarning
            )
    else:
        using_roi = False
        if len(args.annotation_files) == 0:
            printer.write("Either an ROI file or at least annotation file must be given.")
            sys.exit(1)
        else:
            warnings.warn(
                "Using a transcript annotation file instead of an ROI file can lead to double-counting of codons if the annotation contains multiple transcripts per gene.",
                ArgumentWarning
            )
            regions = an.get_transcripts_from_args(args, printer=printer)
            back_buffer = -codon_buffer
            transform_fn = lambda x: x.get_cds()

    phase_sums = {}
    for k in read_lengths:
        phase_sums[k] = numpy.zeros(3)

    for n, roi in enumerate(regions):
        if n % 1000 == 1:
            printer.write("Counted %s ROIs ..." % n)

        # transformation needed to extract CDS from transcript or from ROI file window
        cds_part = transform_fn(roi)

        # only calculate for coding genes
        if len(cds_part) > 0:

            read_dict = {}
            count_vectors = {}
            for k in read_lengths:
                read_dict[k] = []
                count_vectors[k] = []

            # for each seg, fetch reads, sort them, and create individual count vectors
            for seg in cds_part:
                reads = gnd.get_reads(seg)
                for read in filter(lambda x: len(x.positions) in read_dict, reads):
                    read_dict[len(read.positions)].append(read)

                # map and sort by length
                for read_length in read_dict:
                    count_vector = list(gnd.map_fn(read_dict[read_length], seg)[1])
                    count_vectors[read_length].extend(count_vector)

            # add each count vector for each length to total
            for k, vec in count_vectors.items():
                counts = numpy.array(vec)
                if cds_part.strand == "-":
                    counts = counts[::-1]

                if len(counts) % 3 == 0:
                    counts = counts.reshape((int(len(counts) / 3), 3))
                else:
                    if using_roi == False:
                        message = "Length of '%s' coding region (%s nt) is not divisible by 3. Ignoring last partial codon." % (
                            roi.get_name(), len(counts)
                        )
                        warnings.warn(message, DataWarning)
                    newlen = int(len(counts) // 3)
                    counts = counts[:3 * newlen]
                    counts = counts.reshape(newlen, 3)

                phase_sums[k] += counts[codon_buffer:back_buffer, :].sum(0)

    printer.write("Counted %s ROIs total." % (n + 1))
    for k in dtmp:
        dtmp[k] = numpy.array(dtmp[k])

    # total reads counted for each size
    for k in read_lengths:
        dtmp["reads_counted"][dtmp["read_length"] == k] = phase_sums[k].sum()

    # read length distribution
    dtmp["fraction_reads_counted"
         ] = dtmp["reads_counted"].astype(float) / dtmp["reads_counted"].sum()

    # phase vectors
    phase_vectors = {K: V.astype(float) / V.astype(float).sum() for K, V in phase_sums.items()}
    for i in range(3):
        dtmp["phase%s" % i] = numpy.zeros(len(dtmp["read_length"]))

    for k, vec in phase_vectors.items():
        for i in range(3):
            dtmp["phase%s" % i][dtmp["read_length"] == k] = vec[i]

    # phase table
    fn = "%s_phasing.txt" % args.outbase
    printer.write("Saving phasing table to %s ..." % fn)
    dtmp = pd.DataFrame(dtmp)
    with argsopener(fn, args) as fh:
        dtmp.to_csv(
            fh,
            columns=[
                "read_length",
                "reads_counted",
                "fraction_reads_counted",
                "phase0",
                "phase1",
                "phase2",
            ],
            float_format="%.6f",
            na_rep="nan",
            sep="\t",
            index=False,
            header=True
        )
        fh.close()

    fig = {}
    if args.figsize is not None:
        fig["figsize"] = tuple(args.figsize)

    colors = pp.get_colors_from_args(args, len(read_lengths))

    fn = "%s_phasing.%s" % (args.outbase, args.figformat)
    printer.write("Plotting to %s ..." % fn)
    plot_counts = numpy.vstack([V for (_, V) in sorted(phase_sums.items())])
    fig, (ax1,_) = phase_plot(plot_counts,labels=read_lengths,lighten_by=0.3,
                                cmap=None,color=colors,fig=fig)

    if args.title is not None:
        ax1.set_title(args.title)
    else:
        ax1.set_title("Phasing stats for %s" % args.outbase)

    fig.savefig(fn, dpi=args.dpi, bbox_inches="tight")


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""This module provides a thin compatibility layer between Python 2.7 and 3.x.
Various objects are aliased as follows:

    ===================    ===============================   =======================
    **Exported object**     **Points to in 2.x**              **Points to 3.x**
    -------------------    -------------------------------   -----------------------
    ``safe_bytes``         :func:`bytes`                      ``lambda x: bytes(x,"utf-8")``
    ``safe_str``           :func:`str`                        ``lambda x: x.decode("utf-8")``
    ``cStringIO``          :mod:`cStringIO`                   :mod:`io`
    ``StringIO``           :mod:`StringIO`                    :mod:`io`
    ``xrange``             :func:`xrange`                     :func:`range`
    ``quote``              :func:`urllib.quote`               :func:`urllib.parse.quote`
    ``unquote``            :func:`urllib.quote`               :func:`urllib.parse.unquote`
    ``quote_plus``         :func:`urllib.quote_plus`          :func:`urllib.parse.quote_plus`
    ``unquote_plus``       :func:`urllib.unquote_plus`        :func:`urllib.parse.unquote_plus`
    ``ifilter``            :func:`itertools.ifilter`          :func:`filter`
    ``Iterable``           :class:`collections.Iterable`      :class:`collections.abc.Iterable`
    ===================    ===============================   =======================


Also, one function is defined:

    ========================    ================================================
    **Function**                **Action**
    ------------------------    ------------------------------------------------
    :func:`get_func_code`       Retrieves ``function.func_code`` in 2.x, ``function.__code__`` in 3.x
    ========================    ================================================

"""
import sys
import urllib
import itertools
import collections

# yapf: disable
if sys.version_info >= (3, ):
    import io as cStringIO
    import io as StringIO
    import io
    xrange       = range
    quote        = urllib.parse.quote
    unquote      = urllib.parse.unquote
    quote_plus   = urllib.parse.quote_plus
    unquote_plus = urllib.parse.unquote_plus
    ifilter      = filter
    safe_bytes   = lambda x: bytes(x, "utf-8")
    safe_str     = lambda x: x.decode("utf-8")
    safe_file    = io.FileIO

    # function code
    _func_code_attr = "__code__"

    # abstract base classes
    Iterable = collections.abc.Iterable
    
else:
    import cStringIO
    import StringIO
    xrange       = xrange
    quote        = urllib.quote
    unquote      = urllib.unquote
    quote_plus   = urllib.quote_plus
    unquote_plus = urllib.unquote_plus
    ifilter      = itertools.ifilter
    safe_bytes   = bytes
    safe_str     = str
    safe_file    =   file

    # function code
    _func_code_attr = "func_code"

    # abstract base classes
    Iterable = collections.Iterable


def get_func_code(func):
    """Return function code attribute for function ``func``
    
        ==========     ===================
        Version        Function attribute
        ----------     -------------------
        2.x            ``func_code``
        3.x             ``_code``
        ==========     ===================


    Parameters
    ----------
    func : function
        Query function
    
    Returns
    -------
    code
        Function code
    """
    return getattr(func, _func_code_attr)



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.bin.phase_by_size`"""
import tempfile
import os

from nose.plugins.attrib import attr
from pkg_resources import resource_filename, cleanup_resources
from plastid.util.services.decorators import catch_stderr
from plastid.test.functional.base import execute_helper
from plastid.test.ref_files import REF_FILES, COUNT_OPTIONS
from plastid.bin.phase_by_size import main

#===============================================================================
# INDEX: global constants
#===============================================================================

test_info = {
    "test_method": catch_stderr()(main),
    "module_name": "plastid.bin.phase_by_size",
    "ref_file_path": resource_filename("plastid", "test/data/command_line"),
    "temp_file_path": tempfile.mkdtemp(prefix="phase_by_size"),
}

# Define tests as tuples of:
# 1. Command-line style arguments to pass to ``main``
# 2. A list of reference files that output should be compared against
# 3. A list of output files created by running ``main`` with the arguments provided in (1)
# 4. A list of strings specifying how equality should be evaluated
_basename = os.path.join(test_info["temp_file_path"], "test_phase")

#===============================================================================
# INDEX: tests
#===============================================================================

# old test, on annotation file
phase_by_size_tests = [
    # without ROI file
    (
        "%s --min_length 26 --max_length 31 --annotation_files %s --annotation_format BED %s" %
        (_basename, REF_FILES["yeast_mini_bed"], COUNT_OPTIONS), [REF_FILES["yeast_phasing"]],
        [_basename + "_phasing.txt"], [""]
    ),
    # with ROI file
    (
        "%s %s --min_length 26 --max_length 31 %s" %
        (REF_FILES["yeast_metagene_cds_start"], _basename, COUNT_OPTIONS),
        [REF_FILES["yeast_phasing_roifile"]], [_basename + "_phasing.txt"], [""]
    ),
]
"""Functional tests of :py:mod:`plastid.bin.phase_by_size`.

Tests are specified as tuples of:

    1. Command-line style arguments to pass to :py:func:`main`

    2. A list of reference files that output should be compared against

    3. A list of output files created by running :py:func:`main`
       with the arguments provided in (1)

    4. A list of strings specifying how equality should be evaluated
"""

#===============================================================================
# INDEX: test functions
#===============================================================================


@attr(test="functional")
@attr(speed="slow")
def do_test():
    """Perform functional test for :py:mod:`plastid.bin.phase_by_size`"""
    for x in execute_helper(test_info, phase_by_size_tests):
        yield x



#!/usr/bin/env python
"""Function decorators useful for scripts or analyses

Decorators
----------
:py:func:`catch_warnings`
    Catch warnings raised by wrapped function
    
:py:func:`deprecated`
    Function or class decorator. Wrapped functions or classes raise FutureWarnings
    when called or instantiated, respectively

:py:func:`parallelize`
    Parallelize the running of a single-parameter function over multiple instances
    of its parameter

:py:func:`catch_stdout`
    Redirect standard out from a wrapped function into a buffer

:py:func:`catch_stderr`
    Redirect standard error from a wrapped function into a buffer

:py:func:`in_separate_process`
    Run decorated function in a separate process, to force garbage collection
    of that function's memory contents when the function completes and reduce
    overall long-term memory usage.

:py:func:`notimplemented`
    Wrapped functions raise NotImplementedErrors. Use if committing incomplete code.

:py:func:`notused`
    No effects. Used for code annotation only

:py:func:`skip_if_abstract`
    Function decorator for unit tests. Wrapped methods will be skipped if
    they are called from a :py:class:`unittest.TestCase` with `'Abstract'`
    in its name, and run only in subclasses of the abstract :py:class:`unittest.TestCase`
    in which they are defined
"""
import functools
import os
import sys
import warnings
import types
from plastid.util.services.mini2to3 import get_func_code
from plastid.util.services.exceptions import warn_explicit_onceperfamily


def notimplemented(func):
    """NotImplemented annotation decorator.
    Calls to functions annotated with this decorator raise
    :class:`NotImplementedError`, which record attributes of function callers

    Parameters
    ----------
    func : function
        Function to wrap

    Returns
    -------
    function
        wrapped function
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        func_code = get_func_code(func)
        message = "NotImplementedException: call to unimplemented "+\
                  "function %s in module %s line %s" % (func.__name__,
                                                        func.__module__,
                                                        func_code.co_firstlineno + 1)

        raise NotImplementedError(message)

    return new_func


def notused(func):
    """Notused annotation decorator. Only for marking code.

    Parameters
    ----------
    func : function
        Function to wrap

    Returns
    -------
    function
        wrapped function
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        return func(*args, **kwargs)

    return new_func


def catch_warnings(simple_filter="ignore"):
    """Function factory producing function decorators that suppress warnings
    
    Parameters
    ----------
    simple_filter : str
        Warnings filter action. Quoted from :py:mod:`warnings`:
        
            ==============    ==================================================
            **Value**         **Disposition**
            --------------    --------------------------------------------------
            *error*           Turn warnings into exceptions
            *ignore*          Ignore all warnings
            *always*          Always print warnings
            *default*         Print first occurrence of each warning type,
                              for each location where warning is issued
            *module*          Print first occurrence of each warning type,
                              for each module where warning is issued
            *once*            Print first occurrence of matching warnings,
                              regardless of location
            ==============    ==================================================

        (source: :py:mod:`warnings`)
        
        This decorator does NOT support the `onceperfamily` custom warning action.

    
    Returns
    -------
    function
        Decorator function

        
    See also
    --------
    warnings
        Warnings module, especially sections on warnings filters
    """

    def decorator(func):
        """Function decorator that catches warnings of type %s
        
        Parameters
        ----------
        func : function
            Function to decorate
        
        Returns
        -------
        function
            Wrapped function
        """ % simple_filter

        @functools.wraps(func)
        def new_func(*args, **kwargs):
            with warnings.catch_warnings():
                warnings.simplefilter(simple_filter)
                result = func(*args, **kwargs)

            return result

        return new_func

    return decorator


def deprecated(func=None, version=None, instead=None):
    """Deprecation annotation decorator for functions or classes. Wrapped
    functions or classes will raise FutureWarnings with called or instantiated,
    respectively.
    
    Parameters
    ----------
    func : function or class, optional
        Function or class to deprecate. This is just a hack to main backward
        compatibility with non-keyword accepting versions of `deprecated`.

    version : str, optional
        If not `None`, version by which deprecated function or class will be removed
    
    instead : str, optional
        If not `None`, users will be told to use this function or class instead 

    Returns
    -------
    object
        wrapped function or class
    """

    def decorator(func_or_class):
        message = "is deprecated and will be removed from module %s in " % func_or_class.__module__
        if version is not None:
            message += ("plastid version %s." % version)
        else:
            message += "future versions of plastid."

        if instead is not None:
            message += " Use %s instead." % instead

        message = "%s '%s' " + message
        # Based on useful hints from http://wiki.python.org/moin/PythonDecoratorLibrary
        if isinstance(func_or_class, types.FunctionType):
            message = message % ("Function", func_or_class.__name__)

            @functools.wraps(func_or_class)
            def new_func(*args, **kwargs):
                func_code = get_func_code(func_or_class)
                warn_explicit_onceperfamily(
                    message,
                    category=FutureWarning,
                    filename=func_code.co_filename,
                    lineno=func_code.co_firstlineno + 1
                )
                return func_or_class(*args, **kwargs)

            return new_func
        # two tests here. Top line for Python 2.7. Bottom for 3.x
        elif (sys.version_info <= (3,) and (isinstance(func_or_class,types.ClassType) or isinstance(func_or_class,types.TypeType))) \
              or isinstance(func_or_class,type):
            old_init = func_or_class.__init__
            message = message % ("Class", func_or_class.__name__)

            @functools.wraps(func_or_class.__init__)
            def new_func(self, *args, **kwargs):
                warn_explicit_onceperfamily(
                    message,
                    category=FutureWarning,
                    filename=sys._getframe(1).f_globals["__name__"],
                    lineno=sys._getframe(1).f_lineno
                )
                return old_init(self, *args, **kwargs)

            func_or_class.__init__ = new_func
            return func_or_class
        else:
            raise TypeError(
                "Attempt to deprecate '%s', a %s. Only functions and classes can be deprecated." %
                (func_or_class.__name__, type(func_or_class))
            )

    if func is not None:
        return decorator(func)

    return decorator


def skip_if_abstract(func):
    """Decorator function to keep :py:mod:`unittest` from running methods
    (defined in abstract classes) that are only intended to be run when
    inherited by fully-fleshed out subclasses. Wrapped methods are actually
    called from all non-abstract subclasses that inherit the method.
    
    Parameters
    ----------
    func : function
        Function that should only be run in a non-abstract subclass

    Returns
    -------
    function
        wrapped function
    """
    import unittest

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        if "Abstract" in args[0].__class__.__name__:
            return unittest.skip(
                "Skipping all tests from abstract class (don't worry, this is expected)."
            )(
                func
            )
        else:
            return func(*args, **kwargs)

    return new_func


def catch_stderr(buf=None):
    """Function factory producing decorators that capture stderr to a buffer
    
    Parameters
    ----------
    buf : file
        Buffer that will hold captured stderr output. Must import
        ``write()`` and ``fileno()`` methods. If `None`, :py:obj:`os.devnull` will
        be used.
        
        
    Examples
    --------
    Create a ``pipe``, and use it to catch stderr::
    
        >>> import sys
        >>> import os
        >>>
        >>> def my_func():
        >>>    sys.stderr.write("some message")
        
        >>> read_end, write_end = os.pipe()
        >>> buf = os.fdopen(write_end,"w")

        >>> wrapped = catch_stderr(buf)(my_func)
        >>> wrapped() # generates stderr
    
        >>> buf.close()
        >>> captured = os.fdopen(read_end).read()
        >>> print(captured)
        # output from captured stderr here
        
        >>> captured.close() # remember to close!
    
    Returns
    -------
    function
        Function decorator
    """

    def decorator(func, buf=buf):
        """Decorator that suppresses standard error output from a function
        
        Parameters
        ----------
        func : function
            Function whose standard error will be captured
        
        Returns
        -------
        function
            wrapped function    
        """
        if buf is None:
            buf = open(os.devnull, "a")

        @functools.wraps(func)
        def new_func(*args, **kwargs):
            stderr_fd = os.dup(sys.__stderr__.fileno())
            new_fd = buf.fileno()
            os.dup2(new_fd, sys.stderr.fileno())
            try:
                result = func(*args, **kwargs)
            except BaseException as e:
                sys.stderr.flush()
                os.dup2(stderr_fd, sys.stderr.fileno())
                raise (e)

            sys.stderr.flush()
            os.dup2(stderr_fd, sys.stderr.fileno())
            return result

        return new_func

    return decorator


# cannot write unit tests for this because nose substitutes
# a StringIO object for sys.stdout, which messes up the file descriptor
# properties
def catch_stdout(buf=None):
    """Function factory producing decorators that capture stdout to a buffer
    
    Parameters
    ----------
    buf : file or None
        Buffer that will hold captured stdout output. Must import
        ``write()`` and ``fileno()`` methods. If `None`, :py:obj:`os.devnull` will
        be used.
        
        
    Examples
    --------
    Create a ``pipe``, and use it to catch stdout::
    
        >>> import sys
        >>> import os
        >>>
        >>> def my_func():
        >>>     sys.stdout.write("some message")
        
        >>> read_end, write_end = os.pipe()
        >>> buf = os.fdopen(write_end,"w")
        >>> 
        >>> wrapped = catch_stdout(buf)(my_func)
        >>> wrapped() # generates stdout
    
        >>> buf.close()
        >>> captured = os.fdopen(read_end).read()
        >>> print(captured)
        # output here
        
        >>> captured.close() # remember to close!
    
    Returns
    -------
    function
        Function decorator
    """

    def decorator(func, buf=buf):
        """Decorator that suppresses standard error output from a function
        
        Parameters
        ----------
        func : function
            Function whose standard error will be captured
        
        buf : file-like, optional
            Open stream, which **must** have a `fileno()` method. ``StringIO``
            objects will not work! (Default: `os.devnull`)
        
        Returns
        -------
        function
            wrapped function    
        """
        if buf is None:
            buf = open(os.devnull, "a")

        @functools.wraps(func)
        def new_func(*args, **kwargs):
            stdout_fd = os.dup(sys.stdout.fileno())
            new_fd = buf.fileno()
            os.dup2(new_fd, sys.stdout.fileno())
            try:
                result = func(*args, **kwargs)
            except BaseException as e:
                sys.stdout.flush()
                os.dup2(stdout_fd, sys.stdout.fileno())
                raise (e)

            sys.stdout.flush()
            os.dup2(stdout_fd, sys.stdout.fileno())
            return result

        return new_func

    return decorator


def in_separate_process(func):
    """Decorator that runs a function in a separate process, to force garbage collection
    collection upon termination of that process, limiting long-term memory usage.
    
    Parameters
    ----------
    func : function
        Function to run in a separate process. Per multiprocessing spec,
        must be declared in global scope.
    
    Returns
    -------
    function
        wrapped function
    
    See Also
    --------
    multiprocessing
        Python multiprocessing module, for writing parallel programs
    """

    @functools.wraps(func)
    def new_func(*args, **kwargs):
        import multiprocessing
        recv_pipe, send_pipe = multiprocessing.Pipe(False)

        def temp_func(conn, args, kwargs):
            result = func(*args, **kwargs)
            conn.send(result)
            conn.close()

        proc = multiprocessing.Process(target=temp_func, args=(send_pipe, args, kwargs))
        proc.start()
        proc.join()
        result = recv_pipe.recv()
        recv_pipe.close()

        return result

    return new_func


def parallelize(func):
    """Decorator to parallelize the running of a single-parameter function
    over multiple instances of its parameter(s)
     
    Parameters
    ----------
    func : function
        Function to parallelize. Must take one parameter. Per multiprocessing spec,
        must be declared in global scope.
     
    processes : int, optional
        Number of processes to use (Default: `4`)
     
    Returns
    -------
    function
        wrapped function
     
    See Also
    --------
    multiprocessing
        Python multiprocessing module, for writing parallel programs
    """

    @functools.wraps(func)
    def new_func(args, processes=4, chunksize=None, **kwargs):
        import multiprocessing
        pf = functools.partial(func, **kwargs)
        pool = multiprocessing.Pool(processes=processes)
        pool_results = pool.map(pf, args, chunksize=chunksize)
        pool.close()
        pool.join()

        return pool_results

    message = """
     
    Notes
    -----
    #. This function has been parallelized with %s processes via
       :py:func:`plastid.util.services.decorators.parallelize`. Therefore,
       supply a lists of the function's parameter, rather than a single parameter.
    
    #. Output is not guaranteed to be sorted.
    
    #. This function additionally takes the keyword argument ``processes``,
       which determines how many processes it will use.
    """

    if new_func.__doc__ is not None:
        new_func.__doc__ += message
    else:
        new_func.__doc__ = message
    return new_func


def skipdoc(func_or_class):
    """Instruct Sphinx not to generate documentation for ``func_or_class``

    Parameters
    ----------
    func_or_class : function or class
        Function or class not to document

    Returns
    -------
    object
        Wrapped function or class
    """
    func_or_class.plastid_skipdoc = True
    return func_or_class



"""This module contains tools for lookup of features in a region of interest within the a genome.

.. contents::
   :local:

Summary
-------

It is frequently useful to retrieve features that overlap specific regions 
of interest in the genome. ``GenomeHashes`` index features by location,
providing quick lookup.


Module contents
---------------

Several implementations are provided, depending how the data are formatted:

====================    =========================================================
**Implementation**      **Format of feature data**
--------------------    ---------------------------------------------------------
|GenomeHash|            Objects in memory or in unindexed `BED`_, `GTF2`_,
                        `GFF3`_, or `PSL`_ files

|BigBedGenomeHash|      Annotations in `BigBed`_ files

|TabixGenomeHash|       Annotations in `tabix`_-compressed `BED`_, `GTF2`_,
                        `GFF3`_, or `PSL`_ files
====================    =========================================================


Examples
--------
Create a |GenomeHash|::

    >>> from plastid import *

    # from objects in memory
    >>> one_hash = GenomeHash(list_of_transcripts)
    
    # from a non-indexed file
    >>> my_hash = GenomeHash(list(GFF3_Reader("some_file.gff")))
    
    # from a BigBed file
    >>> bigbed_hash = BigBedGenomeHash("some_file.bb")
    
    # from tabix-compressed BED file
    >>> tabix_hash = TabixGenomeHash("some_file.bed.gz","BED")


To find features overlapping a region of interest, pass the feature coordinates
to a |GenomeHash| as a |GenomicSegment|, |SegmentChain|, or |Transcript|::

    >>> overlapping = my_hash[GenomicSegment("chrII",50,10000,"+")]
    >>> overlapping
    [ list of SegmentChains / Transcripts, et c ]
    
    # SegmentChains & Transcripts can also be keys:
    >>> tx = Transcript(GenomicSegment("chrII",50,300,"+"),GenomicSegment("chrII",9000,10000,"+"))
    >>> overlapping2 = my_hash[tx]
    >>> overlapping2
    [ list of SegmentChains / Transcripts, et c ]
    
    # find features that overlap `roi` on either strand
    >>> either_strand_overlap = my_hash.get_overlapping_features(roi,stranded=False)
    
"""

#===============================================================================
# Memory-efficient ways to hash features across a genome
#===============================================================================
import copy
from plastid.util.services.mini2to3 import cStringIO
from plastid.util.io.openers import NullWriter, multiopen
from plastid.readers.bed import BED_Reader
from plastid.readers.gff import GTF2_Reader, GFF3_Reader
from plastid.readers.psl import PSL_Reader
from plastid.genomics.roitools import GenomicSegment, SegmentChain
from abc import abstractmethod

DEFAULT_BIN_SIZE = 20000


class AbstractGenomeHash(object):
    """Abstract base class for objects that index |SegmentChains| by genomic position,
    allowing quick lookup for comparisons of overlap or other behavior
    """

    @abstractmethod
    def get_overlapping_features(self, roi, stranded=True):
        """Return list of features that overlap `roi`
        
        Parameters
        ----------
        feature : |GenomicSegment| or |SegmentChain|
            Query feature

        stranded : bool
            if `True`, retrieve only features on same strand as query feature.
            Otherwise, retrieve features on both strands (Default: `True`)
                             
                             
        Returns
        -------
        list
            List of overlapping features.


        Raises
        ------
        TypeError
            if `roi` is not a |GenomicSegment| or |SegmentChain|
        """
        pass

    @abstractmethod
    def __getitem__(self, roi):
        """Return list of features that overlap a region of interest (`roi`)
        on the same strand
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Query feature indicating region of interest

                             
        Returns
        -------
        list
            Overlapping features


        Raises
        ------
        TypeError
            if feature is not a |GenomicSegment| or |SegmentChain|
        """
        pass


class GenomeHash(AbstractGenomeHash):
    """Index memory-resident features (e.g. |SegmentChains| or |Transcripts|) by genomic position for quick lookup later.
        
     Parameters
     ----------
     features : dict or list, optional
         dict or list of features, as |SegmentChain| objects or subclasses
         (Default: `[]`)
        
     binsize : int, optional
         Size in nucleotides of *neighborhood* for hash. (Default: %s)
        
     do_copy : bool
         If `True`, features will be copied before being stored
         in the hash. This comes at a speed cost, but will prevent unexpected
         side effects if the features are being changed outside the hash. 
         
         If `False` (default), creation of the |GenomeHash| will be much faster.   
         
          
    Notes
    -----
    Because all features are stored in memory, for large genomes, a |TabixGenomeHash|
    or |BigBedGenomeHash| is much more memory-efficient. 
    """

    def __init__(self, features=None, binsize=DEFAULT_BIN_SIZE, do_copy=False):
        """Create a |GenomeHash|
        
         Parameters
         ----------
         features : dict or list, optional
             dict or list of features, as |SegmentChain| objects or subclasses
             (Default: `[]`)
            
         binsize : int, optional
             Size in nucleotides of *neighborhood* for hash. (Default: %s)
            
         do_copy : bool
             If `True`, features will be copied before being stored
             in the hash. This comes at a speed cost, but will prevent unexpected
             side effects if the features are being changed outside the hash. 
             
             If `False` (default), creation of the |GenomeHash| will be much faster.
        """ % DEFAULT_BIN_SIZE
        self.copy = do_copy
        self.feature_dict = {}
        self._id_to_names = {}
        self.binsize = binsize
        features = [] if features is None else features
        self.update(features)

    def __repr__(self):
        return "<%s features=%s binsize=%s chrs=%s>" % (
            self.__class__.__name__, len(self.feature_dict), self.binsize, self._feature_hash.keys()
        )

    def __str__(self):
        return repr(self)

    def update(self, features):
        """Add features to the |GenomeHash|
        
        Parameters
        ----------
         features : dict or list
             dict or list of features, as |SegmentChain| objects or subclasses
        """
        try:
            start = 1 + max(self._id_to_names.keys())
        except ValueError:
            # if empty dictionary
            start = 0

        if isinstance(features, dict):
            for n, (feature_name, feature) in enumerate(features.items()):
                feature_id = start + n
                self._id_to_names[feature_id] = feature_name
                if self.copy == True:
                    self.feature_dict[feature_id] = copy.deepcopy(feature)
                else:
                    self.feature_dict[feature_id] = feature
        else:
            for n, feature in enumerate(features):
                feature_id = start + n
                feature_name = feature.get_name()
                self._id_to_names[feature_id] = feature_name
                if self.copy == True:
                    self.feature_dict[feature_id] = copy.deepcopy(feature)
                else:
                    self.feature_dict[feature_id] = feature

        self._feature_hash = self._make_hash()

    def _make_hash(self):
        """Create the |GenomeHash|
           
        Returns
        -------
        dict
            Hierarchichal dictionary: `dict[chrom][strand] = list<str>`
        """
        my_hash = {}
        for feature_id, feature in self.feature_dict.items():
            bins = self._get_hash_bins(feature)
            chrom = feature.spanning_segment.chrom
            strand = feature.spanning_segment.strand
            if chrom not in my_hash.keys():
                my_hash[chrom] = {}
                my_hash[chrom]["+"] = {}
                my_hash[chrom]["-"] = {}
            for b in bins:
                try:
                    my_hash[chrom][strand][b].append(feature_id)
                except KeyError:
                    my_hash[chrom][strand][b] = [feature_id]
        return my_hash

    def _get_hash_bins(self, roi):
        """Returns a list of genome bins that a given roi falls into
        
        Parameters
        ----------
        feature : |GenomicSegment| or |SegmentChain|
            Query feature
           
           
        Returns
        -------
        list<int>
            List of hash bins that `roi` covers
               
           
        Raises
        ------
        TypeError
            if `roi` is not a |GenomicSegment| or |SegmentChain|
        """
        if isinstance(roi, GenomicSegment):
            rois = [roi]
        elif isinstance(roi, SegmentChain):
            rois = roi
        else:
            raise TypeError("Query feature must be a GenomicSegment or SegmentChain")
        bins = []
        for seg in rois:
            bin_start = seg.start // self.binsize
            bin_end = seg.end // self.binsize
            bins.extend(range(bin_start, bin_end + 1))

        return list(set(bins))

    def _get_nearby_feature_ids(self, roi, stranded=True):
        """Return unique IDs of |SegmentChain| s in all the bins occupied by `roi`
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
           Query feature
           
        stranded : bool
            If `True`, retrieve only features on same strand as query feature.
            Otherwise, retrieve features on both strands
           
                             
        Returns
        -------
        list<str>
           Unique IDs of features near (within *self.binsize* distance from) `roi`
           
           
        Raises
        ------
        TypeError
            if `roi` is not |GenomicSegment| or |SegmentChain|
        """
        if isinstance(roi, GenomicSegment):
            iv = roi
        elif isinstance(roi, SegmentChain):
            iv = roi.spanning_segment
        else:
            raise TypeError("Query feature must be a GenomicSegment or SegmentChain")
        feature_ranges = self._get_hash_bins(roi)
        nearby_feature_ids = []
        try:
            nearby_hash = self._feature_hash[iv.chrom][iv.strand]
        except KeyError:
            nearby_hash = {}

        for b in feature_ranges:
            try:
                nearby_feature_ids.extend(nearby_hash[b])
            except KeyError:
                pass

        if stranded is False:
            other_strand = "+" if iv.strand == "-" else "-" if iv.strand == "+" else "."
            try:
                nearby_hash = self._feature_hash[iv.chrom][other_strand]
            except KeyError:
                nearby_hash = {}

            for b in feature_ranges:
                try:
                    nearby_feature_ids.extend(nearby_hash[b])
                except KeyError:
                    pass

        nearby_feature_ids = set(nearby_feature_ids)
        return nearby_feature_ids

    def get_nearby_features(self, roi, stranded=True):
        """Return list of features in all the bins occupied by `roi`
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Query feature

        stranded : bool
            if `True`, retrieve only features on same strand as query feature.
            Otherwise, retrieve features on both strands
                             
                             
        Returns
        -------
        list<str>
           Features near (within `self.binsize` distance from) `roi`
               

        Raises
        ------
        TypeError
            if `roi` is not a |GenomicSegment| or |SegmentChain|
        """
        nearby_feature_ids = self._get_nearby_feature_ids(roi, stranded=stranded)
        return [self.feature_dict[X] for X in nearby_feature_ids]

    def get_nearby_feature_names(self, roi, stranded=True):
        """Return list of the names of features in all the bins occupied by `roi`
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Query feature

        stranded : bool
            if `True`, retrieve only features on same strand as query feature.
            Otherwise, retrieve features on both strands
                             
                             
        Returns
        -------
        list
           Names of features near (within `self.binsize` distance from) `roi`

        Raises
        ------
        TypeError
            if `roi` is not a |GenomicSegment| or |SegmentChain|
        """
        nearby_feature_ids = self._get_nearby_feature_ids(roi, stranded=stranded)
        return [self._id_to_names[X] for X in nearby_feature_ids]

    def get_overlapping_features(self, roi, stranded=True):
        """Return list of features overlapping `roi`.
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Query feature indicating region of interest

        stranded : bool
            if `True`, retrieve only features on same strand as query feature.
            Otherwise, retrieve features on both strands
                             
                             
        Returns
        -------
        list
           Features overlapping `roi`

        Raises
        ------
        TypeError
            if `roi` is not a |GenomicSegment| or |SegmentChain|
        """
        nearby_features = self.get_nearby_features(roi, stranded=stranded)
        if isinstance(roi, GenomicSegment):
            roi = SegmentChain(roi)

        if stranded == False:
            fn = roi.unstranded_overlaps
        else:
            fn = roi.overlaps
        return [X for X in nearby_features if fn(X) == True]

    def __getitem__(self, roi):
        """Return list of features that overlap a region of interest (roi),
        on same strand.
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Query feature indicating region of interest

        stranded : bool
            if `True`, retrieve only features on same strand as query feature.
            Otherwise, retrieve features on both strands
                             
                             
        Returns
        -------
        list
           Features overlapping `roi`


        Raises
        ------
        TypeError
            if `roi` is not a |GenomicSegment| |SegmentChain|
        """
        return self.get_overlapping_features(roi, stranded=True)


class BigBedGenomeHash(AbstractGenomeHash):
    """BigBedGenomeHash(*filenames,return_type=SegmentChain)
    
    Find features overlapping query regions in `BigBed`_ files.
    
        
    Parameters
    ----------
    *filenames : str 
        One or more filenames to open (NOT open filehandles)

    return_type : class implementing a :py:meth:`from_bed` method
        Class of object to return (Default: |SegmentChain|)
        
    
    Attributes
    ----------
    bigbedreaders : |BigBedReader|
       |BigBedReaders| connecting to BigBed file(s) 
    """

    def __init__(self, *filenames, **kwargs):
        """Create a |BigBedGenomeHash|
        
        Parameters
        ----------
        *filenames : str 
            One or more filenames to open (NOT open filehandles)

        return_type : class implementing a :py:meth:`from_bed` method
            Class of object to return (Default: |SegmentChain|)
        """
        from plastid.readers.bigbed import BigBedReader
        return_type = kwargs.get("return_type", SegmentChain)

        filenames = list(multiopen(filenames))
        for filename in filenames:
            if not isinstance(filename, str):
                raise ValueError("`filename` must be a 'str'. Found a '%s'." % type(filename))

        self.filenames = filenames
        self.bigbedreaders = [BigBedReader(X, return_type=return_type) for X in filenames]

    def get_overlapping_features(self, roi, stranded=True):
        """Return list of features overlapping `roi`
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Query feature indicating region of interest

        stranded : bool
            If `True`, retrieve only features on same strand as query feature.
            Otherwise, retrieve features on both strands
                             
                             
        Returns
        -------
        list
            Features that overlap `roi`

        Raises
        ------
        TypeError
            if `roi` is not a |GenomicSegment| or |SegmentChain|
        """
        ltmp = []
        for reader in self.bigbedreaders:
            ltmp.extend(reader.get(roi, stranded=stranded))

        return ltmp

    def __getitem__(self, roi, stranded=True):
        """Return list of features that overlap the region of interest (roi)
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Query feature indicating region of interest

        stranded : bool
            If `True`, retrieve only features on same strand as query feature.
            Otherwise, retrieve features on both strands
                             
                             
        Returns
        -------
        list
            Features that overlap `roi`

        Raises
        ------
        TypeError
            if `roi` is not a |GenomicSegment| or |SegmentChain|
        """
        return self.get_overlapping_features(roi, stranded=stranded)


class TabixGenomeHash(AbstractGenomeHash):
    """
    TabixGenomeHash(*filenames,data_format='GTF2')
    
    Find features overlapping query regions in `Tabix`_-indexed files.
        
    Parameters
    ----------
    filenames : str or list of str
        Filename or list of filenames of `Tabix`_-compressed files

    data_format : str
        Format of tabix-compressed file(s). Choices are:
        `'GTF2'`,`'GFF3'`,`'BED'`,`'PSL'` (Default: `GTF2`)
    
    
    Attributes
    ----------
    filenames : str or list
        Name of file to open or list of filenames to open (NOT open filehandles)
        
    tabix_readers : list of :py:class:`pysam.Tabixfile`
       `Pysam`_ interfaces to underlying data files 
    """

    _READERS = {
        "GTF2": GTF2_Reader,
        "GFF3": GFF3_Reader,
        "BED": BED_Reader,
        "PSL": PSL_Reader,
    }

    def __init__(self, *filenames, **kwargs):
        """Create a |BigBedGenomeHash|
        
        Parameters
        ----------
        filenames : str or list of str
            Filename or list of filenames of `Tabix`_-compressed files

        data_format : str
            Format of tabix-compressed file(s). Choices are:
            `'GTF2'`,`'GFF3'`,`'BED'`,`'PSL'` (Default: `GTF2`)
        """
        from pysam import Tabixfile
        if len(filenames) == 1 and isinstance(filenames[0], list):
            filenames = filenames[0]

        self.filenames = list(multiopen(filenames))
        self.printer = kwargs.get("printer", NullWriter())
        data_format = kwargs.get("data_format", "GTF2")
        try:
            self._reader_class = TabixGenomeHash._READERS[data_format]
        except ValueError:
            msg = "Supported file formats for TabixGenomeHash are: %s" % ", ".join(
                sorted(TabixGenomeHash._READERS.keys())
            )
            self.printer.write(msg)
            raise ValueError(msg)

        self.tabix_readers = [Tabixfile(X) for X in self.filenames]

    def __del__(self):
        try:
            for reader in self.tabix_readers:
                try:
                    reader.close()
                except:
                    pass
        except:
            pass

    def get_overlapping_features(self, roi, stranded=True):
        """Return list of features overlapping `roi`
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Query feature indicating region of interest

        stranded : bool
            If `True`, retrieve only features on same strand as query feature.
            Otherwise, retrieve features on both strands
                             
        Returns
        -------
        list
           Features that overlap `roi`

        Raises
        ------
        TypeError
            if `roi` is not a |GenomicSegment| or |SegmentChain|
        """
        return list(self.__getitem__(roi, stranded=stranded))

    def __getitem__(self, roi, stranded=True):
        """Return list of features that overlap the region of interest (roi).
        
        Parameters
        ----------
        roi : |GenomicSegment| or |SegmentChain|
            Query feature indicating region of interest

        stranded : bool
            If `True`, retrieve only features on same strand as query feature.
            Otherwise, retrieve features on both strands
                             
                             
        Returns
        -------
        list
            Features that overlap `roi`

        Raises
        ------
        TypeError
            if `roi` is not a |GenomicSegment| or |SegmentChain|
        """
        if isinstance(roi, GenomicSegment):
            roi_seg = roi
            roi_chain = SegmentChain(roi)
        elif isinstance(roi, SegmentChain):
            roi_chain = roi
            roi_seg = roi.spanning_segment
        else:
            raise TypeError("Query feature must be a GenomicSegment or SegmentChain")

        chrom = roi_seg.chrom
        feature_text = "\n".join(["\n".join(list(R.fetch(chrom,
                                                         X.start,
                                                         X.end))) \
                                                         for R in self.tabix_readers \
                                                         for X in roi_chain])

        features = (self._reader_class(cStringIO.StringIO(feature_text)))
        if stranded == True:
            features = [X for X in features if roi_chain.overlaps(X)]
        else:
            features = [X for X in features if roi_chain.unstranded_overlaps(X)]
        return features



#!/usr/bin/env python
"""
"""
import unittest
import copy
import numpy
import pandas as pd
from random import shuffle
from pkg_resources import resource_filename, cleanup_resources
from nose.plugins.attrib import attr
from plastid.genomics.roitools import GenomicSegment

# slight hack to keep imported method from being run as a test
# can't use unittest.skip, or else no tests will never be run!
from plastid.bin.test_table_equality import test_dataframe_equality as checkeq
checkeq.__name__ = "checkeq"
checkeq.__module__ = "checkeq"

# components we will use in equality tests
size = 5000


@attr(test="unit")
class TestTestDataframeEquality(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.cols = {
         "intA"   : numpy.random.randint(0, high=2**16, size=size),
         "intB"   : numpy.random.randint(-10, high=20, size=size),
         "idxA"   : numpy.arange(size),
         "chrA"   : numpy.array([chr(65+(X%(91-65))) for X in range(size)]),
         "strA"   : numpy.array([str(GenomicSegment("chrA", X, X+500, "+")) for X in range(size)]),
         "strB"   : numpy.array([str(GenomicSegment("chrB", X/2, X/2+500, "-")) for X in range(size)]),
         "floatA" : 10*numpy.random.randn(size) + 500,
         "floatB" : (10**-5)*numpy.random.random(size),
         "objA"   : numpy.tile(None, 5000),
         "objB"   : numpy.array([GenomicSegment("chrC", X, X+Y, "+") for X, Y in zip(range(size), numpy.random.randint(2, high=1000, size=size))]),
       } # yapf: disable

    def test_dataframe_equality_when_identical(self):
        df1 = pd.DataFrame(self.cols)
        self.assertTrue(checkeq(df1, df1))

    def test_dataframe_equality_no_sort(self):
        df1 = pd.DataFrame(self.cols)
        df2 = pd.DataFrame(copy.deepcopy(self.cols))
        self.assertTrue(checkeq(df1, df2))
        self.assertTrue(checkeq(df2, df1))

    def test_dataframe_equality_within_tol(self):
        tol = 10**-8
        noiseA = tol / 10**2 * numpy.random.randn(size)
        noiseB = tol / 10**2 * numpy.random.randn(size)
        df1 = pd.DataFrame(self.cols)
        df2 = copy.deepcopy(df1)
        df2["floatA"] += noiseA
        df2["floatB"] -= noiseB
        self.assertTrue(checkeq(df1, df2, tol=tol))
        self.assertTrue(checkeq(df2, df1, tol=tol))

    def test_dataframe_inequality_above_tol(self):
        tol = 10**-8
        noiseA = tol * 10**2 * numpy.random.randn(size)
        noiseB = tol * 10**2 * numpy.random.randn(size)
        df1 = pd.DataFrame(self.cols)
        df2 = copy.deepcopy(df1)
        df2["floatA"] += noiseA
        df2["floatB"] -= noiseB
        self.assertFalse(checkeq(df1, df2, tol=tol))
        self.assertFalse(checkeq(df2, df1, tol=tol))

    def test_dataframe_inequality_wrong_columns(self):
        df1 = pd.DataFrame(self.cols)
        df2 = pd.DataFrame({K: copy.deepcopy(self.cols[K]) for K in sorted(self.cols.keys())[:-2]})
        self.assertFalse(checkeq(df1, df2))
        self.assertFalse(checkeq(df2, df1))

    def test_dataframe_inequality_wrong_rows(self):
        df1 = pd.DataFrame(self.cols)
        df2 = pd.DataFrame({K: self.cols[K][:size - 1000] for K in self.cols.keys()})
        self.assertFalse(checkeq(df1, df2))
        self.assertFalse(checkeq(df2, df1))

    def check_dataframe_equality_same(self, special_val):
        """Helper function for checking equality when various special values
        are in the same location in both dataframes

        Parameters
        ----------
        special_val : numpy.nan, numpy.inf, -numpy.inf, or None
            Value that is ignored by :py:meth:`plastid.bin.test_dataframe_equality` if it occurs in the same cells of both dataframes
        """
        idx = numpy.random.randint(0, high=size, size=500)
        tmpcols = copy.deepcopy(self.cols)
        tmpcols["floatA"][idx] = special_val
        df1 = pd.DataFrame(tmpcols)
        df2 = pd.DataFrame(copy.deepcopy(tmpcols))
        self.assertTrue(checkeq(df1, df2))
        self.assertTrue(checkeq(df2, df1))

    def test_dataframe_equality_same_nans(self):
        self.check_dataframe_equality_same(numpy.nan)

    def test_dataframe_equality_same_infs(self):
        self.check_dataframe_equality_same(numpy.inf)
        self.check_dataframe_equality_same(-numpy.inf)

    def check_dataframe_inequality_different(self, special_val):
        """Helper function for checking inequality when various special values
        appear in the different locations in both dataframes

        Parameters
        ----------
        special_val : numpy.nan, numpy.inf, -numpy.inf, or None
            Value that is ignored by :py:meth:`plastid.bin.test_dataframe_equality` if it occurs in the same cells of both dataframes
        """
        idxA = numpy.random.randint(0, high=size, size=500)
        idxB = numpy.random.randint(0, high=size, size=500)

        colsA = copy.deepcopy(self.cols)
        colsA["floatA"][idxA] = special_val

        colsB = copy.deepcopy(self.cols)
        colsB["floatA"][idxB] = special_val

        df1 = pd.DataFrame(colsA)
        df2 = pd.DataFrame(colsB)
        self.assertFalse(checkeq(df1, df2))
        self.assertFalse(checkeq(df2, df1))

    def test_dataframe_inequality_different_nans(self):
        self.check_dataframe_inequality_different(numpy.nan)

    def test_dataframe_inequailty_different_infs(self):
        self.check_dataframe_inequality_different(numpy.inf)
        self.check_dataframe_inequality_different(-numpy.inf)

    def test_dataframe_equality_with_sort_str(self):
        shuffidx = numpy.arange(size)
        shuffle(shuffidx)
        df1 = pd.DataFrame(self.cols)
        df2 = pd.DataFrame({K: self.cols[K][shuffidx] for K in self.cols.keys()})
        self.assertTrue(checkeq(df1, df2, sort_columns=["strA"]))
        self.assertTrue(checkeq(df2, df1, sort_columns=["strA"]))

    def test_dataframe_equality_with_sort_numeric(self):
        shuffidx = numpy.arange(size)
        shuffle(shuffidx)
        df1 = pd.DataFrame(self.cols)
        df2 = pd.DataFrame({K: self.cols[K][shuffidx] for K in self.cols.keys()})
        self.assertTrue(checkeq(df1, df2, sort_columns=["idxA"]))
        self.assertTrue(checkeq(df2, df1, sort_columns=["idxA"]))

    def test_dataframe_equality_with_multi_sort(self):
        shuffidx = numpy.arange(size)
        shuffle(shuffidx)
        df1 = pd.DataFrame(self.cols)
        df2 = pd.DataFrame({K: self.cols[K][shuffidx] for K in self.cols.keys()})
        self.assertTrue(checkeq(df1, df2, sort_columns=["strB", "chrA"]))
        self.assertTrue(checkeq(df2, df1, sort_columns=["strB", "chrA"]))



#!/usr/bin/env python
"""Regression-testing script designed to test equality between a newly-generated
file and a reference file that is intended to contain the same data. Rows and
columns are not expected to be in the same order. Float values are only required
to be equal within a user-specified tolerance. `NaN` values evaluate as equal
if and only if they occur in the same cell after sorting rows and columns.
Finally, specific columns may be excluded by name (or number, if there is no
header row).

Exit status is 0 if files are identical, 1 otherwise
"""
import argparse
import inspect

import sys
import numpy
import pandas as pd
from plastid.util.io.openers import opener, get_short_name, NullWriter
from plastid.util.io.filters import NameDateWriter, CommentReader

printer = NameDateWriter(get_short_name(inspect.stack()[-1][1]))

_NUMERIC_DTYPES = "biufc"


def equal_enough(col1, col2, tol=1e-10, printer=NullWriter()):
    """If `col1` and `col2` are both numeric, test that
    all their values are within `tol` of each other. :obj:`numpy.nan` values,
    if present, must be in the same place in each column. Ditto :obj:`numpy.inf`
    values.

    If `col1` and `col2` are not numeric, return true if they have the same
    `dtype` and the same values in all cells.

    Parameters
    ----------
    col1 : numpy.ndarray
        First column of data

    col2 : numpy.ndarray
        Second column of data

    tol : float
        Error tolerance for numeric data between col1 and col2, value-wise

    printer : anything implementing a ``write()`` method (e.g. a |NameDateWriter|)
        if not None, rich comparison information will be sent to this writer

    Returns
    -------
    bool
        `True` if `col1 == col2` for non-numeric data;
        `True` if `abs(col1 - col2) <= tol` for numeric data;
        `False` otherwise
    """
    dtype_test = col1.dtype.kind in _NUMERIC_DTYPES
    dtype_test &= col2.dtype.kind in _NUMERIC_DTYPES

    if dtype_test:
        nan1 = numpy.isnan(col1)
        nan2 = numpy.isnan(col2)

        inf1 = numpy.isinf(col1)
        inf2 = numpy.isinf(col2)

        # check nans in same place
        nan_test = (nan1 == nan2).all()

        # make sure infs are in same place, and are same sign
        inf_test = (inf1 == inf2).all() & (col1[inf1] == col2[inf1]).all()

        diff_test = (abs(col1[~nan1 & ~inf1] - col2[~nan1 & ~inf1]) <= tol).all()

        if not nan_test:
            printer.write("Failed nan test")
        if not inf_test:
            printer.write("Failed inf test")
        if not diff_test:
            printer.write("Failed tolerance test")

        return nan_test & diff_test & inf_test
    elif col1.dtype.kind not in _NUMERIC_DTYPES and col2.dtype.kind not in _NUMERIC_DTYPES:
        diff_test = (col1 == col2)
        if not diff_test.all():
            ltmp = [
                "%s, %s, %s" % (N, val1, val2) for N,
                (val1, val2) in enumerate(zip(col1[~diff_test], col2[~diff_test]))
            ]
            printer.write("Differences:\n%s" % "\n".join(ltmp))
        return diff_test.all()
    else:
        return False


def test_dataframe_equality(
        df1,
        df2,
        tol            = 1e-8,
        sort_columns   = [],
        printer        = NullWriter(),
        print_verbose  = False,
        return_verbose = False
): # yapf: disable
    """Test equality of dataframes over multiple columns, with verbose output.
    If `NaNs` or `Infs` are present, these must be present in corresponding cells
    in both dataframes for the dataframes to evaluate as equal.

    Parameters
    ----------
    df1 : :py:class:`pd.DataFrame`
        First dataframe

    df2 : :py:class:`pd.DataFrame`
        Second dataframe

    tol : float, optoinal
        Maximum tolerated difference between floating point numbers
        (Default: *1e-8*)

    sort_columns : list, optional
        List of column names or indices on which to sort data before comparing values

    printer : file-like, optional
        Any logger importing a ``write()`` method

    print_verbose : bool, optional
        Print verbose output to stderr (Default: False)

    return_verbose : bool, optional
        If *True*, return a list of failure messages


    Returns
    -------
    bool
        `True` if dataframes are equal, `False` otherwise

    list
        A list of strings explaining how `df1` and `df2` differ. Only
        returned if `return_verbose` is `True`
    """
    if len(sort_columns) > 0:
        df1 = df1.sort_index(axis=1).sort_values(sort_columns, axis=0)
        df2 = df2.sort_index(axis=1).sort_values(sort_columns, axis=0)
    else:
        df1 = df1.sort_values(list(df1.columns), axis=0).sort_index()
        df2 = df2.sort_values(list(df2.columns), axis=0).sort_index()

    failures = []
    keys1 = set(df1.columns)
    keys2 = set(df2.columns)
    retval = True
    if keys1 != keys2:
        failstrs = [
            "Tables contain different columns (unique keys shown below):",
            "    %s: %s" % (1, ",  ".join([str(X) for X in keys1 - keys2])),
            "    %s: %s" % (2, ",  ".join([str(X) for X in keys2 - keys1]))
        ]
        printer.write("\n".join(failstrs))
        failures.extend(failstrs)
        retval = False

    if len(df1) != len(df2):
        failstrs = [
            "Tables contain different numbers of rows:",
            "    %s: %s" % (1, len(df1)),
            "    %s: %s" % (2, len(df2))
        ]
        printer.write("\n".join(failstrs))
        failures.extend(failstrs)
        retval = False

    if retval == True:
        if print_verbose == True:
            printer.write("Sorting data...")

        if print_verbose == True:
            printer.write(
                "Testing equality of values by column with numerical tolerance %.3e..." % tol
            )

        unequal = []
        for k in keys1:
            if print_verbose == True:
                printer.write("    Testing column %s..." % k)
            # take out values to get sorted numpy arrays, avoiding weird reordering
            # done behind the scenes by pandas
            s1 = df1[k].values
            s2 = df2[k].values
            if not equal_enough(s1, s2, tol=tol, printer=printer):
                unequal.append(k)
            else:
                if print_verbose == True:
                    printer.write("        %s is identical" % k)

        failstr = "Column values are unequal for columns: %s" % (
            ", ".join([str(X) for X in sorted(unequal)])
        )
        if len(unequal) > 0:
            printer.write(failstr)
            failures.append(failstr)
            retval = False

    if return_verbose == True:
        return retval, failures
    else:
        return retval


def main(argv=sys.argv[1:], verbose=False):
    """Command-line program

    Parameters
    ----------
    argv : list, optional
        A list of command-line arguments, which will be processed
        as if the script were called from the command line if
        :py:func:`main` is called directly.

        Default: `sys.argv[1:]`. The command-line arguments, if the script is
        invoked from the command line

    verbose : bool, optional
        If `True`, return


    Returns
    -------
    int
        `0` if files are identical, `1` otherwise

    str
        Only returned if `verbose` is selected. String describing how
        tables are unequal (e.g. which columns failed, et c).
    """
    parser = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.add_argument("file1", type=str)
    parser.add_argument("file2", type=str)
    parser.add_argument(
        "-v", dest="verbose", default=False, action="store_true", help="Give verbose output"
    )
    parser.add_argument(
        "--sort_keys",
        default=None,
        metavar="key",
        nargs="+",
        help=
        "If specified,  values will be sorted by the column(s) corresponding to these name or numbers (0-indexed) before comparison"
    )
    parser.add_argument(
        "--exclude",
        type=str,
        default=[],
        nargs="+",
        metavar="key",
        help="Key or number (0-indexed) of columns to exclude"
    )
    parser.add_argument("--no_header", default=False, action="store_true",
                        help="If specified,  no header row is present. Columns "+\
                             "for all other command-line flags "+\
                             "must be referenced by number (starting at zero) "+\
                             "rather than name,  and will be assumed to be in "+\
                             "the same order in both files.")
    parser.add_argument(
        "--tol",
        type=float,
        default=1e-8,
        help="Tolerance by which floats are allowed to differ (Default: 1e-8)"
    )

    args = parser.parse_args(argv)

    kwargs = {
        "sep": "\t",
        "index_col": args.sort_keys,
        "comment": "#",
    }
    exclude = args.exclude
    if args.no_header is True:
        if args.sort_keys is not None:
            kwargs["index_col"] = [int(X) for X in args.sort_keys]
        exclude = [int(X) for X in exclude]

        with opener(args.file1) as fh:
            df1 = pd.read_table(fh, header=None, **kwargs)

        with opener(args.file2) as fh:
            df2 = pd.read_table(fh, header=None, **kwargs)

    else:
        with opener(args.file1) as fh:
            df1 = pd.read_table(fh, header=0, **kwargs)

        with opener(args.file2) as fh:
            df2 = pd.read_table(fh, header=0, **kwargs)

    if len(args.exclude) > 0:
        printer.write("Excluding columns %s: " % ", ".join(args.exclude))

    for k in exclude:
        if k in df1:
            df1.pop(k)
        if k in df2:
            df2.pop(k)

    test_result, messages = test_dataframe_equality(
        df1, df2, printer=printer, print_verbose=True, return_verbose=True, tol=args.tol
    )

    if test_result == True:
        printer.write("Files contain equivalent data.")
        exit_code = 0
    else:
        printer.write("Files non-equivalent.")
        exit_code = 1

    if __name__ == "__main__":
        sys.exit(exit_code)
    else:
        if args.verbose == True or verbose == True:
            return exit_code, messages
        else:
            return exit_code


if __name__ == "__main__":
    main()



#!/usr/bin/env python
"""Test suite for :py:mod:`plastid.bin.reformat_transcripts`"""
import tempfile
import os
import itertools
import shlex
import shutil
from nose.tools import assert_equal, assert_set_equal
from nose.plugins.attrib import attr
from pkg_resources import resource_filename, cleanup_resources
from plastid.readers.bed import BED_Reader
from plastid.readers.gff import GTF2_TranscriptAssembler
from plastid.test.functional.base import execute_helper
from plastid.test.ref_files import REF_FILES, MINI
from plastid.bin.reformat_transcripts import main
from plastid.util.services.decorators import catch_stderr

#===============================================================================
# INDEX: global constants
#===============================================================================

test_info = {
    "test_method": catch_stderr()(main),
    "module_name": "plastid.bin.gff_to_gtf",
    "ref_file_path": resource_filename("plastid", "test/data/command_line"),
    "temp_file_path": tempfile.mkdtemp(prefix="reformat_transcripts"),
}

_INPUT_FORMATS = ["BED", "BigBed", "GTF2", "GFF3"]
_OUTPUT_FORMATS = ["BED", "GTF2"]
_in_out = list(itertools.product(_INPUT_FORMATS, _OUTPUT_FORMATS))
_outfilename = os.path.join(test_info["temp_file_path"], "test_reformat_transcripts")

_test_args = {
    "GTF2":
    " --exclude 5 8 --sort_keys 0 3 4",  # exclude score and attr, which are not guaranteed to be sorted; sort on region
    "BED": " --sort_keys 3",  # sort on region
}

_reader_funcs = {
    "GTF2": (GTF2_TranscriptAssembler,
             {
                 "add_three_for_stop": False
             }),
    "BED": (BED_Reader, {})
}

_default_scores = {
    ("BED", "BED"): 0.0,
    ("BED", "GTF2"): 0.0,
    ("GTF2", "BED"): 0.0,
    ("GTF2", "GTF2"): ".",
    ("GFF3", "BED"): 0.0,
    ("GFF3", "GTF2"): ".",
    ("BigBed", "BED"): 0.0,
    ("BigBed", "GTF2"): 0.0,
}


def check_files_applying_sort(ref_file, test_file, infmt, outfmt):
    opener, args = _reader_funcs[outfmt]
    with open(ref_file) as fh:
        ref_transcripts = sorted(list(opener(fh, **args)))
    with open(test_file) as fh:
        test_transcripts = sorted(list(opener(fh, **args)))

    assert_equal(
        len(ref_transcripts), len(test_transcripts),
        "%s to %s: Length mismatch in discovered transcripts. Expected '%s'. Found '%s'" %
        (infmt, outfmt, len(ref_transcripts), len(test_transcripts))
    )
    for tx1, tx2 in zip(ref_transcripts, test_transcripts):
        assert_equal(
            tx1.get_name(), tx2.get_name(),
            "%s to %s: Found unordered transcripts. Expected '%s'. Found '%s"
            '' % (infmt, outfmt, tx1.get_name(), tx2.get_name())
        )

        set1 = tx1.get_position_set()
        set2 = tx2.get_position_set()
        assert_set_equal(
            set1, set2, "%s to %s: Difference in position sets. Expected '%s'. Found '%s'" %
            (infmt, outfmt, set1, set2)
        )

        ref_score = str(_default_scores[(infmt, outfmt)])
        found_score = str(tx2.attr["score"])
        assert_equal(
            found_score, ref_score,
            "%s to %s: Did not find expected score. Expected: '%s'. Found '%s'" %
            (infmt, outfmt, ref_score, found_score)
        )

        # BED preserves fewer attributes than GTF, so we can only test on common keys
        # We exclude "gene_id" for BEd and BigBed input because this will not match
        # We exclude "score" because we already tested it
        #
        # by testing attr we are also implicitly testing cds_genome_end and cds_genome_start
        attr1 = tx1.attr
        attr2 = tx2.attr
        keyset = set(attr1.keys()) & set(attr2.keys()) - {"score"}
        if infmt in ("BED", "BigBed") and outfmt != "BED":
            keyset -= {"gene_id"}

        for k in keyset:
            assert_equal(
                attr1[k], attr2[k],
                "%s to %s: Difference in attribute %s. Expected '%s'. Found '%s'" %
                (infmt, outfmt, k, attr1[k], attr2[k])
            )
        #assert_dict_equal(attr1,attr2,"%s to %s: Difference in attributes. Expected %s. Found %s" % (infmt,outfmt,attr1,attr2))


@attr(test="functional")
@attr(speed="slow")
def do_test():
    """Perform functional test for :py:mod`plastid.bin.reformat_transcripts`"""
    for infmt, outfmt in _in_out:

        my_outfilename = _outfilename + (
            "_%s_to_%s.%s" % (infmt.lower(), outfmt.lower(), outfmt.lower())
        )
        infile = REF_FILES["100transcripts_%s" % infmt.lower().replace("2", "").replace("3", "")]
        ref_outfile = REF_FILES["100transcripts_%s" % outfmt.lower().replace("2",
                                                                             "").replace("3", "")]
        argstr = "%s --annotation_files %s --annotation_format %s --output_format %s" % (
            my_outfilename, infile, infmt, outfmt
        )
        test_info["test_method"](shlex.split(argstr))
        yield check_files_applying_sort, ref_outfile, my_outfilename, infmt, outfmt


def tearDownModule():
    if test_info["temp_file_path"] != "":
        shutil.rmtree(test_info["temp_file_path"])



#!/usr/bin/env python
"""This module contains parsers for data structures written in the `autoSql`_
object specification language, used by the `UCSC genome browser`_, `BigBed`_ files
and `BigWig`_ files.

.. contents::
   :local:
   
   
Summary
-------

Parsers are constructed by initializing an |AutoSqlDeclaration| with a block of
`autoSql`_ text::

    >>> declaration = '''table easy_table
    "A table with a comment on the next line" 
        (
        uint number auto; "a number with a token"
        uint [3] points ; "r,g,b values"
        lstring  my_string ; "a long string"
        uint a_field_size ; "the size for the next field"
        float [a_field_size] float_array ; "an array of floats"
        set(a,b,c) alpha ; "the first three letters of the alphabet"
        )
    '''
    >>> record_parser = AutoSqlDeclaration(declaration)


The parser that is created can then be called to parse text records into dictionaries::

    >>> record_parser("3    1,2,3    my string with spaces    5    1.1,1.2,1.3,1.4,1.5    a,b")
    OrderedDict([("number",3),
                 ("points",(1,2,3)),
                 ("my_string","my string with spaces"),
                 ("a_field_size",5),
                 ("float_array",(1.1,1.2,1.3,1.4,1.5)),
                 ("alpha",{'a','b'}]))

Module contents
---------------
|AutoSqlDeclaration|
    Parses `autoSql`_ declarations for `table`, `simple`, and `object`
    declaration types. Delegates parsing of individual fields to appropriate subclasses 
    (e.g. |AutoSqlField|, |SizedAutoSqlField|, and |ValuesAutoSqlField|).

|AutoSqlField|, |SizedAutoSqlField|, |ValuesAutoSqlField|
    Parse various sorts of fields within an autoSql declaration block


Notes
-----
 #. These parsers seek only to provide Python bindings for `autoSql`_ declarations.
    They do **NOT** generate C or SQL code from `autoSql`_, as those functions
    are already provided by `Jim Kent's utilities <https://github.com/ENCODE-DCC/kentUtils/tree/master/>`_

 #. ``set`` and ``enum`` field types are parsed as ``sets`` of strings

 #. ``primary``, ``index``, and ``auto`` `autoSQL`_ tags are accepted in line declarations,
    but are ignored because they are not relevant for parsing

 #. The parsers assume that they will be parsing tab-delimited text blocks
  
 #. Although declarations are routinely nested as fields within other
    declarations in C ``struct`` s and in SQL databases, in the absence of a standard,
    it is unclear how these would be serialized within tab-delimited columns of `BigBed`_
    files. Therefore, nested declarations are not supported..


See Also
--------
`Updated autoSql Grammar specification <https://github.com/ENCODE-DCC/kentUtils/blob/36d6274459f644d5400843b8fa097b380b8f7867/src/hg/autoSql/autoSql.doc>`_
    Explanation of autoSql grammar

`The ENCODE project's tests for autoSql parsers <https://github.com/ENCODE-DCC/kentUtils/tree/master/src/hg/autoSql/tests/input>`_
    Official autoSql unit tests

`Kent & Brumbaugh, 2002 <http://www.linuxjournal.com/article/5949>`_
    First publication of autoSql & autoXml 

"""
import re
from collections import OrderedDict
from abc import abstractmethod
from plastid.util.services.exceptions import DataWarning, warn, filterwarnings

# use "once" because we want each literal to be shown once
filterwarnings("once", "Could not find formatter for field.*")
filterwarnings("once", "Could not find formatter for field.*")
filterwarnings("once", "Could not convert autoSql value.*")
filterwarnings("once", ".*already found in autoSql declaration.*")

# regular expressions that recognize various autoSql elements
# TODO: migrate this to Parsimonious or something more sane
_pattern_bits = { 
    "start"     : r"^\s*",
    "type"      : r"(?P<type>\w+)",
    "name"      : r"\s+(?P<name>\w+)\s*",
    "semi"      : r"\s*;\s*",
    "comment"   : r"\"(?P<comment>[^\"]*)\"",
    "size"      : r"\s*\[\s*(?P<size>\w+)\s*\]\s*",
    "values"    : r"\s*\(\s*(?P<value_names>[^()]+)\s*\)\s*",
    "optionals" : r"(?P<opt1>\s+primary|\s+auto|\s+index\s*(\[\s*\d+\s*\])?)?\s*(?P<opt2>\s+primary|\s+auto|\s+index\s*(\[\s*\d+\s*\])?)?\s*(?P<opt3>\s+primary|\s+auto|\s+index\s*(\[\s*\d+\s*\])?)?", 
    "declare_type_name" : r"(?P<declare_type>object|simple|table)\s+(?P<declare_name>\w+)\s+",
    "field_text" :  r"\s*\(\s*(?P<field_text>.*)\)",
 } # yapf: disable


class AbstractAutoSqlElement(object):
    """Abstract base class for parsers of autoSql elements
    
    Attributes
    ----------
    attr : dict
        Dictionary of attributes describing the element (e.g. *name,* *type,* et c) 
        
    autosql : str
        Block of autoSql text specifying format of element
        
    match_pattern : :py:class:`re.RegexObject`
        Pattern that determines whether or not a block of autoSql matches this object
    
    parent : instance of subclass of |AbstractAutoSqlElement|, or None
        Parent / enclosing element
    
    field_types : dict
        Dictionary matching type names (as strings) to formatters that parse them
        from plaintext
    
    delim : str, optional
        Text delimiter for fields in blocks called by :py:meth:~__call__~
        (Default: "\t")
    """
    match_str = ""
    match_pattern = re.compile(match_str)

    def __init__(self, autosql, parent=None, delim="\t"):
        self.autosql = autosql
        self.parent = parent
        self.delim = delim
        self.field_types = {
            "int"     : (int, "i"),   #32-bit
            "uint"    : (int, "I"),   #32-bit
            "short"   : (int, "h"),   #16-bit
            "ushort"  : (int, "H"),   #16-bit
            "byte"    : (int, "b"),   #8-bit
            "ubyte"   : (int, "B"),   #8-bit
            "float"   : (float, "f"), #single-precision
            "char"    : (str, "c"),   #8-bit
            "string"  : (str, "s"),   #variable up to 255bytes
            "lstring" : (str, "s"),   #variable up to 2billion bytes
        } # yapf: disable
        self.attr = self.match_pattern.search(autosql).groupdict()

    def __repr__(self):
        return "<%s name=%s type=%s>" % (
            self.__class__.__name__, self.attr["name"],
            self.attr.get("type", self.__class__.__name__)
        )

    def add_type(self, name, formatter):
        """Add a type to the parser
        
        Parameters
        ----------
        name : str
            Name of data type
        
        formatter : callable
            Function/callable that, when applied to autoSql text, yields
            an object of the type specified by ``name``
        """
        self.field_types[name] = formatter

    @abstractmethod
    def __call__(self, text, rec=None):
        """Parse an OrderedDict matching ``self.autosql`` from a block of delimited text
        
        Parameters
        ----------
        text : str
            Multiline text block, formatted in autoSql
        
        rec : OrderedDict or None, optional
            Record whose attributes are being populated by recursive
            processing of ``text``. Passed in cases where fields sized by variables
            need to look up instance values of earlier fields to evaluate those
            variables.
        """
        pass

    @staticmethod
    def mask_comments(text):
        """Mask all comments in an autoSql block in order to facilitate parsing
        by regular expressions
        
        Parameters
        ----------
        text : str
            autoSql-formatted text
        
        Returns
        -------
        str
            Text with comments replaced by "xxxxxx" of same length
        
        list
            List of (comment.start,comment.end), including quotes, for each comment
            in ``text`` 
        """
        cpat = re.compile(r"\"[^\"]+\"")
        match_locs = []
        for match in cpat.finditer(text):
            my_start = match.start()
            my_end = match.end()
            match_len = my_end - my_start
            match_locs.append((my_start + 1, my_end - 1))
            text = text[:my_start + 1] + "x" * (match_len - 2) + text[my_end - 1:]

        return text, match_locs

    @classmethod
    def matches(cls, text):
        """Determine whether autoSql formatting text matches this autoSql element
        
        Parameters
        ----------
        text : str
            Block of autoSql-formatted declaration text
        
        Returns
        bool
            True an autoSql parser of this class's type can be made from this
            specification, otherwise False
        """
        return cls.match_pattern.search(text) is not None


class AutoSqlDeclaration(AbstractAutoSqlElement):
    """Parser factory that converts delimited text blocks into OrderedDicts,
    following the field names and types described by an autoSql declaration element

    Parameters
    ----------
    autosql : str
        Block of autoSql text specifying format of element
        
    parent : instance of subclass of |AbstractAutoSqlObject| or `None`, optional
        Parent / enclosing element. Default: None
    
    delim : str, optional
        Field delimiter (default: tab)    
    
    
    Attributes
    ----------
    attr : dict
        Dictionary of descriptive attributes (e.g. *name,* *type,* *declare_type,* et c) 
    
    field_formatters : OrderedDict
        Dictionary mapping field names to type names

    field_comments : OrderedDict
        Dictionary mapping field names to comments
        
    field_types : dict
        Dictionary matching type names (as strings) to formatters that parse them
        from plaintext
    
    autosql : str
        Block of autoSql text specifying format of element
        
    match_pattern : :py:class:`re.RegexObject`
        Pattern that determines whether or not a block of autoSql matches this object
    
    parent : instance of subclass of |AbstractAutoSqlObject|, or None
        Parent / enclosing element. Default: None
    
    delim : str, optional
        Text delimiter for fields in blocks called by :py:meth:~__call__~
        (Default: "\t")
    
    
    Methods
    -------
    :py:meth:`AutoSqlDeclaration.__call__`
        Parse autoSql-formatted blocks of text according to this declaration
    """

    match_str = r"".join(
        [_pattern_bits[_X] for _X in ("start", "declare_type_name", "comment", "field_text")]
    )
    match_pattern = re.compile(match_str, re.S)

    def __init__(self, autosql, parent=None, delim="\n"):
        """Create an |AutoSqlDeclaration|
        
        Parameters
        ----------
        autosql : str
            Block of autoSql text specifying format of element
            
        parent : instance of subclass of |AbstractAutoSqlObject| or None, optional
            Parent / enclosing element. Default: None
        
        delim : str, optional
            Field delimiter (default: tab)
        """
        AbstractAutoSqlElement.__init__(self, autosql, parent=parent, delim="\t")

        # re-do regex match masking out comments, in case the comments
        # contain special characters that would mess up the parsing
        masked_sql, comment_match_locs = self.mask_comments(autosql)
        match_dict = self.match_pattern.search(masked_sql).groupdict()

        self.attr["declare_type"] = match_dict["declare_type"]
        self.attr["name"] = match_dict["declare_name"]
        masked_field_text = match_dict["field_text"]

        self.attr["comment"] = autosql[comment_match_locs[0][0]:comment_match_locs[0][1]
                                       ].strip("\n").strip("\"")
        field_text_start = masked_sql.index(masked_field_text)
        self._field_text = autosql[field_text_start:field_text_start + len(masked_field_text)]

        if self.parent is not None:
            self.parent.add_type(self.attr["declare_name"], self)

        self.field_formatters = OrderedDict()
        self.field_comments = OrderedDict()
        self._parse_fields()

    def _parse_fields(self):
        """Parse fields of an autoSql declaration, and populate
        ``self.field_formatters`` and ``self.field_comments``.
        """
        # order in which we try to match autoSql fields
        match_order = [AutoSqlField, SizedAutoSqlField, ValuesAutoSqlField]

        # fields are area of string from last starting point to end of comment
        # first starting point is 0; all subsequent starting points will be end
        # of previous comment

        _, comment_locs = self.mask_comments(self._field_text)
        last_index = 0
        for (_, next_index) in comment_locs:
            field_str = self._field_text[last_index:next_index + 1]
            for field_class in match_order:
                if field_class.matches(field_str):
                    my_parser = field_class(field_str)
                    name = my_parser.attr["name"]
                    if name in self.field_formatters:
                        oldname = name
                        i = 1
                        current_formatter = self.field_formatters[name]
                        current_type = current_formatter.attr.get(
                            "type", current_formatter.__class__.__name__
                        )
                        new_type = my_parser.attr.get("type", my_parser.__class__.__name__)
                        while name in self.field_formatters:
                            i += 1
                            name = "%s%s" % (oldname, i)
                            warn(
                                "Element named '%s' of type '%s' already found in autoSql declaration '%s.' Renaming current element of type '%s' to '%s'"
                                % (
                                    oldname, current_type,
                                    self.attr.get("name", "unnamed declaration"), new_type, name
                                ), DataWarning
                            )
                        my_parser.attr["name"] = name

                    self.field_formatters[name] = my_parser
                    self.field_comments[name] = my_parser.attr["comment"]

            last_index = next_index + 1

    def __repr__(self):
        return "<%s name=%s type=%s fields=[%s]>" % (
            self.__class__.__name__, self.attr["name"],
            self.attr.get("type", self.__class__.__name__), ",".join(self.field_formatters.keys())
        )

    def __call__(self, text, rec=None):
        """Parse an OrderedDict matching ``self.autosql`` from a block of delimited text
        
        Parameters
        ----------
        text : str
            Multiline text block, formatted in autoSql

        rec : OrderedDict or None, optional
            Record whose attributes are being populated by recursive
            processing of ``text``. Passed in cases where fields sized by variables
            need to look up instance values of earlier fields to evaluate those
            variables.
        
        Returns
        -------
        OrderedDict
            Dictionary mapping field names to their values
        """
        items = text.split(self.delim)
        rec = OrderedDict() if rec is None else rec
        obj = OrderedDict()
        for item, (field_name, formatter) in zip(items, self.field_formatters.items()):
            obj[field_name] = formatter(item, rec=obj)

        return obj


class AutoSqlField(AbstractAutoSqlElement):
    """Parser factory for autoSql fields of type ``fieldType fieldName ';' comment``

        
    Parameters
    ----------
    autosql : str
        Block of autoSql text specifying format of element
        
    parent : instance of subclass of |AbstractAutoSqlObject| or None, optional
        Parent / enclosing element. Default: None
    
    delim : str, optional
        Field delimiter (default: tab)
        
    
    Attributes
    ----------
    attr : dict
        Dictionary of descriptive attributes (e.g. name, type, et c) 

    formatter : callable
        Callable/function that converts plain text into an object of the correct type
        
    autosql : str
        Block of autoSql text specifying format of element
        
    match_pattern : :class:`re.RegexObject`
        Pattern that determines whether or not a block of autoSql matches this object
    
    parent : instance of subclass of :class:`AbstractAutoSqlObject` or `None`
        Parent / enclosing element (Default: None)
    
    delim : str, optional
        Text delimiter for fields in blocks called by :meth:`__call__`
        (Default: newline)
    """
    match_str = r"".join(
        [_pattern_bits[_X] for _X in ("start", "type", "name", "optionals", "semi", "comment")]
    )
    match_pattern = re.compile(match_str)

    def __init__(self, autosql, parent=None, delim=""):
        """Create an |AutoSqlField|
        
        Parameters
        ----------
        autosql : str
            Block of autoSql text specifying format of element
            
        parent : instance of subclass of |AbstractAutoSqlObject| or None, optional
            Parent / enclosing element. Default: None
        
        delim : str, optional
            Field delimiter (default: tab)
        """
        AbstractAutoSqlElement.__init__(self, autosql, parent=parent, delim=delim)
        type_ = self.attr["type"]
        try:
            self.formatter = self.field_types[type_][0]
        except KeyError:
            try:
                self.formatter = self.parent.field_types[type_][0]
            except:
                self.formatter = str
                warn(
                    "Could not find formatter for field '%s' of type '%s'. Casting to 'string' instead."
                    % (self.attr["name"], type_), DataWarning
                )

    def __call__(self, text, rec=None):
        """Parse an value matching the field described by ``self.autosql``
        from a block of delimited text
        
        Parameters
        ----------
        text : str
            Multiline text block, formatted in autoSql
        
        Returns
        -------
        Value or object of appropriate type
        """
        try:
            return self.formatter(text)
        except ValueError:
            message = "Could not convert autoSql value '%s' for field '%s' to type '%s'. Casting to 'string' instead. " % (
                text, self.attr["name"], self.formatter.__name__
            )
            warn(message, DataWarning)
            return text


class SizedAutoSqlField(AutoSqlField):
    """Parser factory for autoSql fields of type ``fieldType `[` fieldSize `]` fieldName ';' comment``

        
    Parameters
    ----------
    autosql : str
        Block of autoSql text specifying format of element
        
    parent : instance of subclass of |AbstractAutoSqlObject| or None, optional
        Parent / enclosing element. Default: None
    
    delim : str, optional
        Field delimiter (default: tab)


    Attributes
    ----------
    attr : dict
        Dictionary of descriptive attributes (e.g. *name*, *size,* *type,* et c) 

    formatter : callable
        Callable/function that converts plain text into an object of the correct type
        
    autosql : str
        Block of autoSql text specifying format of element
        
    match_pattern : :class:`re.RegexObject`
        Pattern that determines whether or not a block of autoSql matches this object
    
    parent : instance of subclass of :class:`AbstractAutoSqlObject` or `None`
        Parent / enclosing element (Default: None)
    
    delim : str, optional
        Text delimiter for fields in blocks called by :meth:`__call__`
        (Default: newline)

    Methods
    -------
    :py:meth:`SizedAutoSqlField.__call__`
        Parse autoSql-formatted blocks of text into the tuples of the object type
        specified by this field  
    """
    match_str = r"".join(
        [
            _pattern_bits[_X]
            for _X in ("start", "type", "size", "name", "optionals", "semi", "comment")
        ]
    )
    match_pattern = re.compile(match_str)

    def __init__(self, autosql, size=1, parent=None, delim=","):
        """Create a |SizedAutoSqlField|
        
        Parameters
        ----------
        autosql : str
            Block of autoSql text specifying format of element
            
        parent : instance of subclass of |AbstractAutoSqlObject| or None, optional
            Parent / enclosing element. Default: None
        
        delim : str, optional
            Field delimiter (default: tab)
        """
        AutoSqlField.__init__(self, autosql, parent=parent, delim=delim)
        try:
            self.attr["size"] = int(self.attr["size"])
            self.attr["size_is_int"] = True
        except ValueError:
            self.attr["size_is_int"] = False

    def __call__(self, text, rec=None):
        """Parse an value matching the field described by ``self.autosql``
        from a block of delimited text
        
        Parameters
        ----------
        text : str
            Multiline text block, formatted in autoSql

        rec : OrderedDict or None, optional
            Record whose attributes are being populated by recursive
            processing of ``text``. Passed in cases where fields sized by variables
            need to look up instance values of earlier fields to evaluate those
            variables.
        
        Returns
        -------
        tuple
            Tuple of appropriate type
        """
        if self.formatter != str:
            try:
                retval = tuple(
                    [self.formatter(X) for X in text.strip().strip(self.delim).split(self.delim)]
                )
            except ValueError:
                message = "Could not convert autoSql value '%s' in field '%s' to tuple of type '%s'. Leaving as str " % (
                    text, self.attr["name"], self.formatter.__name__
                )
                warn(message, DataWarning)
                return text
        else:
            retval = text

        if self.attr["size_is_int"] == True:
            assert len(retval) == self.attr["size"]
        else:
            assert len(retval) == rec[self.attr["size"]]

        return retval


# for set, enum types
class ValuesAutoSqlField(AbstractAutoSqlElement):
    """Parser factory for autoSql fields of type ``fieldType `(` fieldValues `)` fieldName ';' comment``
    where ``fieldType`` would typically be ``set`` or ``enum``
    
        
    Parameters
    ----------
    autosql : str
        Block of autoSql text specifying format of element
        
    parent : instance of subclass of |AbstractAutoSqlObject| or None, optional
        Parent / enclosing element. Default: None
    
    delim : str, optional
        Field delimiter (default: tab)    
    """

    match_str = r"".join(
        [
            _pattern_bits[_X]
            for _X in ("start", "type", "values", "name", "optionals", "semi", "comment")
        ]
    )
    match_pattern = re.compile(match_str)

    def __init__(self, autosql, parent=None, delim=","):
        """Create a |ValuesAutoSqlField|
        
        Parameters
        ----------
        autosql : str
            Block of autoSql text specifying format of element
            
        parent : instance of subclass of |AbstractAutoSqlObject| or None, optional
            Parent / enclosing element. Default: None
        
        delim : str, optional
            Field delimiter (default: tab)
        """
        AbstractAutoSqlElement.__init__(self, autosql, parent=parent, delim=delim)
        self.attr["value_names"] = [X.strip() for X in self.attr["value_names"].split(",")]

    def __call__(self, text, rec=None):
        """Parse an value matching the field described by ``self.autosql``
        from a block of delimited text
        
        Parameters
        ----------
        text : str
            Multiline text block, formatted in autoSql

        rec : OrderedDict or None, optional
            Record whose attributes are being populated by recursive
            processing of ``text``. Passed in cases where fields sized by variables
            need to look up instance values of earlier fields to evaluate those
            variables.
        
        Returns
        -------
        set
            set of items found in column 
        """
        items = set(
            [X.strip() for X in text.strip(self.delim).split(self.delim) if len(X.strip()) > 0]
        )
        return items
